# SPDX-License-Identifier: GNU GPL v3

"""
Processes of an image by applying filters to it and converting it to a binary version.
"""

import cv2
import numpy as np
from matplotlib import pyplot as plt
from cv2.typing import MatLike
from dataclasses import dataclass
from skimage.morphology import disk
from skimage.filters.rank import autolevel, median

from ..utils.config_loader import load_img_configs
from ..utils.sgt_utils import safe_uint8_image



class BaseImage:
    """
    A class that is used to binarize an image by applying filters to it and converting it to a binary version.

    Args:
        raw_img (MatLike): Raw image in OpenCV format
        scale_factor (float): Scale factor used to downsample/up-sample the image.
    """

    @dataclass
    class ScalingKernel:
        """A data class for storing scaling kernel parameters."""
        image_patches: list[MatLike]
        kernel_shape: tuple
        # stride: tuple

    def __init__(self, raw_img: MatLike|None, cfg_file="", scale_factor=1.0):
        """
        A class that is used to binarize an image by applying filters to it and converting it to a binary version.

        Args:
            raw_img: Raw image in OpenCV format
            cfg_file (str): Configuration file path
            scale_factor (float): Scale factor used to downsample/up-sample the image.
        """
        self._configs: dict = load_img_configs(cfg_file)  # image processing configuration parameters and options.
        self._img_raw: MatLike | None = safe_uint8_image(raw_img)
        self._img_2d: MatLike | None = None
        self._img_bin: MatLike | None = None
        self._img_mod: MatLike | None = None
        self._img_hist: MatLike | None = None
        self._has_alpha_channel: bool = False
        self._scale_factor: float = scale_factor
        self._image_filters: list[BaseImage.ScalingKernel] = []
        self.init_image()

    @property
    def configs(self) -> dict:
        """Returns the image processing configuration parameters and options."""
        return self._configs

    @configs.setter
    def configs(self, configs: dict) -> None:
        """Sets the image processing configuration parameters and options."""
        self._configs = configs

    @property
    def img_raw(self) -> MatLike | None:
        """Returns the raw image in OpenCV format."""
        return self._img_raw

    @property
    def img_2d(self) -> MatLike | None:
        """Returns the processed image in OpenCV format."""
        return self._img_2d

    @img_2d.setter
    def img_2d(self, img_2d: MatLike | None) -> None:
        """Sets the processed image in OpenCV format."""
        self._img_2d = img_2d

    @property
    def img_bin(self) -> MatLike | None:
        """Returns the binary image in OpenCV format."""
        return self._img_bin

    @img_bin.setter
    def img_bin(self, img_bin: MatLike | None) -> None:
        """Sets the binary image in OpenCV format."""
        self._img_bin = img_bin

    @property
    def img_mod(self) -> MatLike | None:
        """Returns the modified image in OpenCV format."""
        return self._img_mod

    @img_mod.setter
    def img_mod(self, img_mod: MatLike | None) -> None:
        """Sets the modified image in OpenCV format."""
        self._img_mod = img_mod

    @property
    def img_hist(self) -> MatLike | None:
        """Returns the histogram of the processed image."""
        return self._img_hist

    @property
    def has_alpha_channel(self) -> bool:
        """Returns whether the image has an alpha channel."""
        return self._has_alpha_channel

    @property
    def scale_factor(self) -> float:
        """Returns the scale factor used to downsample/up-sample the image."""
        return self._scale_factor

    @scale_factor.setter
    def scale_factor(self, scale_factor: float) -> None:
        """Sets the scale factor used to downsample/up-sample the image."""
        self._scale_factor = scale_factor

    @property
    def image_filters(self) -> list["BaseImage.ScalingKernel"]:
        """Returns the list of scaling kernels used to the image."""
        return self._image_filters

    @image_filters.setter
    def image_filters(self, image_filters: list["BaseImage.ScalingKernel"]) -> None:
        """Sets the list of scaling kernels used to the image."""
        self._image_filters = image_filters

    def init_image(self) -> None:
        """
        Initialize the class member variables (or attributes).
        Returns:

        """
        if self._img_raw is None:
            return
        img_data = self._img_raw.copy()

        self._has_alpha_channel, _ = BaseImage.check_alpha_channel(self._img_raw)
        self._img_2d = img_data

    def get_pixel_width(self) -> None:
        """Compute pixel dimension in nanometers to estimate and update the width of graph edges."""

        def compute_pixel_width(scalebar_val: float, scalebar_pixel_count: int) -> float:
            """
            Compute the width of a single pixel in nanometers.

            :param scalebar_val: Unit value of the scale in nanometers.
            :param scalebar_pixel_count: Pixel count of the scalebar width.
            :return: Width of a single pixel in nanometers.
            """

            val_in_meters = scalebar_val / 1e9
            pixel_width = val_in_meters / scalebar_pixel_count
            return pixel_width

        opt_img = self._configs
        pixel_count = int(opt_img["scalebar_pixel_count"]["value"])
        scale_val = float(opt_img["scale_value_nanometers"]["value"])
        if (scale_val > 0) and (pixel_count > 0):
            px_width = compute_pixel_width(scale_val, pixel_count)
            opt_img["pixel_width"]["value"] = px_width / self._scale_factor

    def apply_img_crop(self, x: int, y: int, crop_width: int, crop_height: int, actual_w: int, actual_h: int) -> None:
        """
        A function that crops images into a new box dimension.

        :param x: Left coordinate of cropping box.
        :param y: Top coordinate of cropping box.
        :param crop_width: Width of cropping box.
        :param crop_height: Height of cropping box.
        :param actual_w: Width of actual image.
        :param actual_h: Height of actual image.
        """

        # Resize image
        scaled_img = cv2.resize(self._img_2d.copy(), (actual_w, actual_h))

        # Crop image
        self._img_2d = scaled_img[y:y + crop_height, x:x + crop_width]

    def process_img(self, image: MatLike) -> MatLike | None:
        """
        Apply filters to the image.

        :param image: OpenCV image.
        :return: None
        """

        opt_img = self._configs
        if image is None:
            return None

        def control_brightness(img: MatLike):
            """
            Apply contrast and brightness filters to the image

            param img: OpenCV image
            :return:
            """

            brightness_val = opt_img["brightness_level"]["value"]
            contrast_val = opt_img["contrast_level"]["value"]
            brightness = ((brightness_val / 100) * 127)
            contrast = ((contrast_val / 100) * 127)

            # img = np.int16(img)
            # img = img * (contrast / 127 + 1) - contrast + brightness
            # img = np.clip(img, 0, 255)
            # img = np.uint8(img)

            if brightness != 0:
                if brightness > 0:
                    shadow = brightness
                    max_val = 255
                else:
                    shadow = 0
                    max_val = 255 + brightness
                alpha_b = (max_val - shadow) / 255
                gamma_b = shadow
                img = cv2.addWeighted(img, alpha_b, img, 0, gamma_b)

            if contrast != 0:
                alpha_c = float(131 * (contrast + 127)) / (127 * (131 - contrast))
                gamma_c = 127 * (1 - alpha_c)
                img = cv2.addWeighted(img, alpha_c, img, 0, gamma_c)

            # text string in the image.
            # cv2.putText(new_img, 'B:{},C:{}'.format(brightness, contrast), (10, 30), cv2.FONT_HERSHEY_SIMPLEX,
            # 1, (0, 0, 255), 2)
            return img

        def apply_filter(filter_type: str, img: MatLike, fil_grad_x, fil_grad_y):
            """"""
            if filter_type == 'scharr' or filter_type == 'sobel':
                abs_grad_x = cv2.convertScaleAbs(fil_grad_x)
                abs_grad_y = cv2.convertScaleAbs(fil_grad_y)
                fil_dst = cv2.addWeighted(abs_grad_x, 0.5, abs_grad_y, 0.5, 0)
                fil_abs_dst = cv2.convertScaleAbs(fil_dst)
                result_img = cv2.addWeighted(img, 0.75, fil_abs_dst, 0.25, 0)
                return cv2.convertScaleAbs(result_img)
            return img

        if len(image.shape) == 3:
            image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)

        # Apply brightness/contrast
        filtered_img = control_brightness(image)

        if float(opt_img["apply_gamma"]["dataValue"]) != 1.00:
            inv_gamma = 1.00 / float(opt_img["apply_gamma"]["dataValue"])
            inv_gamma = float(inv_gamma)
            lst_tbl = [((float(i) / 255.0) ** inv_gamma) * 255.0 for i in np.arange(0, 256)]
            table = np.array(lst_tbl).astype('uint8')
            filtered_img = cv2.LUT(filtered_img, table)

        # applies a low-pass filter
        if opt_img["apply_lowpass_filter"]["value"] == 1:
            h, w = filtered_img.shape
            ham1x = np.hamming(w)[:, None]  # 1D hamming
            ham1y = np.hamming(h)[:, None]  # 1D hamming
            ham2d = np.sqrt(np.dot(ham1y, ham1x.T)) ** int(
                opt_img["apply_lowpass_filter"]["dataValue"])  # expand to 2D hamming
            f = cv2.dft(filtered_img.astype(np.float32), flags=cv2.DFT_COMPLEX_OUTPUT)
            f_shifted = np.fft.fftshift(f)
            f_complex = f_shifted[:, :, 0] * 1j + f_shifted[:, :, 1]
            f_filtered = ham2d * f_complex
            f_filtered_shifted = np.fft.fftshift(f_filtered)
            inv_img = np.fft.ifft2(f_filtered_shifted)  # inverse F.T.
            filtered_img = np.abs(inv_img)
            filtered_img -= filtered_img.min()
            filtered_img = filtered_img * 255 / filtered_img.max()
            filtered_img = filtered_img.astype(np.uint8)

        # applying median filter
        if opt_img["apply_median_filter"]["value"] == 1:
            # making a 5x5 array of all 1's for median filter
            med_disk = disk(5)
            filtered_img = median(filtered_img, med_disk)

        # applying gaussian blur
        if opt_img["apply_gaussian_blur"]["value"] == 1:
            b_size = int(opt_img["apply_gaussian_blur"]["dataValue"])
            filtered_img = cv2.GaussianBlur(filtered_img, (b_size, b_size), 0)

        # applying auto-level filter
        if opt_img["apply_autolevel"]["value"] == 1:
            # making a disk for the auto-level filter
            auto_lvl_disk = disk(int(opt_img["apply_autolevel"]["dataValue"]))
            filtered_img = autolevel(filtered_img, footprint=auto_lvl_disk)

        # applying a scharr filter,
        if opt_img["apply_scharr_gradient"]["value"] == 1:
            # applying a scharr filter, and then taking that image and weighting it 25% with the original,
            # this should bring out the edges without separating each "edge" into two separate parallel ones
            d_depth = cv2.CV_16S
            grad_x = cv2.Scharr(filtered_img, d_depth, 1, 0)
            grad_y = cv2.Scharr(filtered_img, d_depth, 0, 1)
            filtered_img = apply_filter('scharr', filtered_img, grad_x, grad_y)

        # applying sobel filter
        if opt_img["apply_sobel_gradient"]["value"] == 1:
            scale = 1
            delta = 0
            d_depth = cv2.CV_16S
            grad_x = cv2.Sobel(filtered_img, d_depth, 1, 0, ksize=int(opt_img["apply_sobel_gradient"]["dataValue"]),
                               scale=scale,
                               delta=delta, borderType=cv2.BORDER_DEFAULT)
            grad_y = cv2.Sobel(filtered_img, d_depth, 0, 1, ksize=int(opt_img["apply_sobel_gradient"]["dataValue"]),
                               scale=scale,
                               delta=delta, borderType=cv2.BORDER_DEFAULT)
            filtered_img = apply_filter('sobel', filtered_img, grad_x, grad_y)

        # applying laplacian filter
        if opt_img["apply_laplacian_gradient"]["value"] == 1:
            d_depth = cv2.CV_16S
            dst = cv2.Laplacian(filtered_img, d_depth, ksize=int(opt_img["apply_laplacian_gradient"]["dataValue"]))
            # dst = cv2.Canny(img_filtered, 100, 200); # canny edge detection test
            abs_dst = cv2.convertScaleAbs(dst)
            filtered_img = cv2.addWeighted(filtered_img, 0.75, abs_dst, 0.25, 0)
            filtered_img = cv2.convertScaleAbs(filtered_img)

        return filtered_img

    def binarize_img(self, image: MatLike) -> MatLike | None:
        """
        Convert image to binary.

        :param image:
        :return: None
        """

        if image is None:
            return None

        img_bin = None
        opt_img = self._configs
        otsu_res = 0  # only needed for the OTSU threshold

        # Applying the universal threshold, checking if it should be inverted (dark foreground)
        if opt_img["threshold_type"]["value"] == 0:
            if opt_img["apply_dark_foreground"]["value"] == 1:
                img_bin = \
                cv2.threshold(image, int(opt_img["global_threshold_value"]["value"]), 255, cv2.THRESH_BINARY_INV)[1]
            else:
                img_bin = cv2.threshold(image, int(opt_img["global_threshold_value"]["value"]), 255, cv2.THRESH_BINARY)[
                    1]

        # adaptive threshold generation
        elif opt_img["threshold_type"]["value"] == 1:
            if self._configs["adaptive_local_threshold_value"]["value"] <= 1:
                # Bug fix (crushes app)
                self._configs["adaptive_local_threshold_value"]["value"] = 3

            if opt_img["apply_dark_foreground"]["value"] == 1:
                img_bin = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                                cv2.THRESH_BINARY_INV,
                                                int(opt_img["adaptive_local_threshold_value"]["value"]), 2)
            else:
                img_bin = cv2.adaptiveThreshold(image, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                                cv2.THRESH_BINARY,
                                                int(opt_img["adaptive_local_threshold_value"]["value"]), 2)

        # OTSU threshold generation
        elif opt_img["threshold_type"]["value"] == 2:
            if opt_img["apply_dark_foreground"]["value"] == 1:
                temp = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY_INV + cv2.THRESH_OTSU)
                img_bin = temp[1]
                otsu_res = temp[0]
            else:
                temp = cv2.threshold(image, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
                img_bin = temp[1]
                otsu_res = temp[0]
        self._configs["otsu"]["value"] = otsu_res
        return img_bin

    def evaluate_img_binary(self) -> tuple[float, np.ndarray] | tuple[None, None]:
        """A function that evaluates the pre-processed image binary by overlaying the binary image on top of the
        original image and masking sections of the image that do not intersect with "white" (255) pixels in the
        binary image. The unmasked sections are typically where generated graph edges and nodes are located. So, the 
        unmasked sections should have fairly the same pixel values in the original image and the binary image. In the 
        binary image the pixel values are 255 "white", while in the original image they are typically 0-255, but with 
        small variations. The Standard Deviation (SD) can help identify how different the pixel values are in the 
        unmasked sections of the original image. Also, a histogram of the pixel values in the unmasked sections of the 
        original image can help identify the distribution of pixel values.
        
        :return: The Standard Deviation and Histogram of the unmasked sections (in the original image).
        """
        
        if self._img_2d is None:
            return None, None
        
        if self._img_bin is None:
            return None, None

        # Find pixel positions where the binary image is white (255)
        white_pixel_pos = np.argwhere(self._img_bin == 255)  # (row, col)

        # Retrieve corresponding pixel values from img_2d
        img_rgb = self._img_2d
        if self._has_alpha_channel:
            img_rgb = self._img_2d[..., :3]
        pixel_values = [img_rgb[tuple(p)] for p in white_pixel_pos]
        pixel_values = np.array(pixel_values)

        # Calculate standard deviation of original values
        std_dev = np.std(pixel_values)

        # Create the histogram of original values at white pixel positions
        eval_hist = cv2.calcHist([pixel_values], [0], None, [256], [0, 256])
        return float(std_dev), eval_hist

    def plot_img_histogram(self, axes=None, curr_view="") -> plt.Figure:
        """
        Uses Matplotlib to plot the histogram of the processed image.

        :param axes: A Matplotlib axes object.
        :param curr_view: The current visualization type of the image (Original, Processed, Binary).
        """
        fig = plt.figure()
        plt_title = "Processed Image"
        if curr_view != "":
            plt_title = f"{curr_view} image"

        if axes is None:
            ax = fig.add_subplot(1, 1, 1)
        else:
            ax = axes
        ax.set(yticks=[], xlabel='Pixel values', ylabel='Counts')
        ax.set_title(plt_title)

        if curr_view == "original":
            img = self._img_2d
            # Evaluate the binary image
            eval_std, eval_hist = self.evaluate_img_binary()
            if eval_std is not None:
                print(f"Evaluating Binary Image (Std. Dev.): {eval_std}")
                ax.plot(eval_hist, color='c', label='Evaluated Binary Histogram')
                ax.legend(loc='upper right')
        elif curr_view == "binary":
            img = self._img_bin
        else:
            img = self._img_mod

        if img is None:
            return fig

        self._img_hist = cv2.calcHist([img], [0], None, [256], [0, 256])
        ax.plot(self._img_hist, label='Image Histogram')
        ax.legend(loc='upper right')
        if self._configs["threshold_type"]["value"] == 0:
            global_val = int(self._configs["global_threshold_value"]["value"])
            thresh_arr = np.array([[global_val, global_val], [0, max(self._img_hist)]], dtype='object')
            ax.plot(thresh_arr[0], thresh_arr[1], ls='--', color='black')
        elif self._configs["threshold_type"]["value"] == 2:
            otsu_val = self._configs["otsu"]["value"]
            thresh_arr = np.array([[otsu_val, otsu_val], [0, max(self._img_hist)]], dtype='object')
            ax.plot(thresh_arr[0], thresh_arr[1], ls='--', color='black')
        fig.tight_layout()
        return fig

    def get_config_info(self) -> str:
        """
        Get the user selected parameters and options information.
        :return:
        """

        opt_img = self._configs

        run_info = "***Image Filter Configurations***\n"
        if opt_img["threshold_type"]["value"] == 0:
            run_info += "Global Threshold (" + str(opt_img["global_threshold_value"]["value"]) + ")"
        elif opt_img["threshold_type"]["value"] == 1:
            run_info += "Adaptive Threshold, " + str(opt_img["adaptive_local_threshold_value"]["value"]) + " bit kernel"
        elif opt_img["threshold_type"]["value"] == 2:
            run_info += "OTSU Threshold"

        if opt_img["apply_gamma"]["value"] == 1:
            run_info += f" || Gamma = {opt_img["apply_gamma"]["dataValue"]}"
        run_info += "\n"
        if opt_img["apply_median_filter"]["value"]:
            run_info += "Median Filter ||"
        if opt_img["apply_gaussian_blur"]["value"]:
            run_info += "Gaussian Blur, " + str(opt_img["apply_gaussian_blur"]["dataValue"]) + " bit kernel || "
        if opt_img["apply_autolevel"]["value"]:
            run_info += "Autolevel, " + str(opt_img["apply_autolevel"]["dataValue"]) + " bit kernel || "
        run_info = run_info[:-3] + '' if run_info.endswith('|| ') else run_info
        run_info += "\n"
        if opt_img["apply_dark_foreground"]["value"]:
            run_info += "Dark Foreground || "
        if opt_img["apply_laplacian_gradient"]["value"]:
            run_info += "Laplacian Gradient || "
        if opt_img["apply_scharr_gradient"]["value"]:
            run_info += "Scharr Gradient || "
        if opt_img["apply_sobel_gradient"]["value"]:
            run_info += "Sobel Gradient || "
        if opt_img["apply_lowpass_filter"]["value"]:
            run_info += "Low-pass filter, " + str(opt_img["apply_lowpass_filter"]["dataValue"]) + " window size || "
        run_info = run_info[:-3] + '' if run_info.endswith('|| ') else run_info
        run_info += "\n\n"

        run_info += "***Microscopy Parameters***\n"
        run_info += f"Scalebar Value = {opt_img["scale_value_nanometers"]["value"]} nm"
        run_info += f" || Scalebar Pixel Count = {opt_img["scalebar_pixel_count"]["value"]}\n"
        run_info += f"Resistivity = {opt_img["resistivity"]["value"]}" + r"$\Omega$m"
        run_info += "\n\n"

        if self._img_raw is not None:
            run_info += "***Image Scale***\n"
            run_info += f"Size = {self._img_2d.shape[0]} x {self._img_2d.shape[1]} px"
            run_info += f" || Scale Factor = {self._scale_factor}"

        return run_info

    @staticmethod
    def check_alpha_channel(img: MatLike) -> tuple[bool, str | None]:
        """
        A function that checks if an image has an Alpha channel or not. Only works for images with up to 4-Dimensions.

        :param img: OpenCV image.
        """

        if img is None:
            return False, None

        if len(img.shape) == 2:
            return False, "Grayscale"

        if len(img.shape) == 3:
            channels = img.shape[2]
            if channels == 4:
                return True, "RGBA"
            elif channels == 3:
                return False, "RGB"
            elif channels == 2:
                return True, "Grayscale + Alpha"
            elif channels == 1:
                return False, "Grayscale"

        # Unknown Format
        return False, None

    @staticmethod
    def resize_img(size: int, image: MatLike) -> tuple[MatLike | None, float | None]:
        """
        Resizes image to specified size.

        :param size: new image pixel size.
        :param image: OpenCV image.
        :return: rescaled image
        """
        if image is None:
            return None, None
        h, w = image.shape[:2]
        if h > w:
            scale_factor = size / h
        else:
            scale_factor = size / w
        std_width = int(scale_factor * w)
        std_height = int(scale_factor * h)
        std_size = (std_width, std_height)
        std_img = cv2.resize(image, std_size)
        return std_img, scale_factor



"""
**StructuralGT**

A software package for performing Graph Theory on microscopic TEM images. This software is a \
modified version of StructuralGT by Drew A. Vecchio: https://github.com/drewvecchio/StructuralGT.

    Copyright (C) 2025, the Regents of the University of Michigan.

        This program is free software: you can redistribute it and/or modify \
it under the terms of the GNU General Public License as published by \
the Free Software Foundation, either version 3 of the License, or \
(at your option) any later version. This program is distributed in \
the hope that it will be useful, but WITHOUT ANY WARRANTY; without \
even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  \
See the GNU General Public License for more details. You should have received a copy \
of the GNU General Public License along with this program.  If not, see <https://www.gnu.org/licenses/>.

        Development Lead: Dickson Owuor

        Contributors: Nicholas A. Kotov

        Contact email: owuordickson@gmail.com
"""


# Project Details
__version__ = "3.6.0"
__install_version__ = "3.5.8"
__title__ = "StructuralGT"
__author__ = "Dickson Owuor"
__credits__ = "The Regents of the University of Michigan"



# SPDX-License-Identifier: GNU GPL v3

"""
Compute graph theory metrics
"""

import os
import math
import time
import datetime
import itertools
import logging
import multiprocessing
import numpy as np
import scipy as sp
import pandas as pd
import igraph as ig
import networkx as nx
import matplotlib.table as tbl
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from collections import defaultdict
from statistics import stdev, StatisticsError
from matplotlib.backends.backend_pdf import PdfPages

from networkx.algorithms.centrality import betweenness_centrality, closeness_centrality, eigenvector_centrality
from networkx.algorithms import average_node_connectivity, global_efficiency, clustering
from networkx.algorithms import degree_assortativity_coefficient
from networkx.algorithms.flow import maximum_flow
from networkx.algorithms.distance_measures import diameter, periphery
from networkx.algorithms.wiener import wiener_index

from ..utils.progress_update import ProgressUpdate
from ..networks.fiber_network import FiberNetworkBuilder
from ..imaging.image_processor import ImageProcessor
from ..utils.config_loader import load_gtc_configs
from ..utils.sgt_utils import get_num_cores, AbortException

logger = logging.getLogger("SGT App")

# WE ARE USING CPU BECAUSE CuPy generates some errors - yet to be resolved.
COMPUTING_DEVICE = "CPU"
"""
try:
    import sys
    import cupy as cp

    # Check for GPU
    test = cp.cuda.Device(0).compute_capability
    # Check for CUDA_PATH in environment variables
    cuda_path = os.getenv("CUDA_PATH")
    print(cuda_path)
    if cuda_path:
        xp = np  # Use CuPy for GPU
        COMPUTING_DEVICE = "GPU"
        logging.info("Using GPU with CuPy!", extra={'user': 'SGT Logs'})
    else:
        logging.info(
            "Please add CUDA_PATH to System environment variables OR install 'NVIDIA GPU Computing Toolkit'\nvia: https://developer.nvidia.com/cuda-downloads",
            extra={'user': 'SGT Logs'})
        raise ImportError("Please add CUDA_PATH to System environment variables.")
except (ImportError, NameError, AttributeError):
    xp = np  # Fallback to NumPy for CPU
    logging.info("Using CPU with NumPy!", extra={'user': 'SGT Logs'})
except cp.cuda.runtime.CUDARuntimeError:
    xp = np  # Fallback to NumPy for CPU
    logging.info("Using CPU with NumPy!", extra={'user': 'SGT Logs'})
"""


def _worker_vertex_connectivity(ig_graph_obj, i, j):
    try:
        lnc = ig_graph_obj.vertex_connectivity(source=i, target=j, neighbors="negative")
        return lnc if lnc != -1 else None
    except Exception as err:
        logging.exception("Computing iGraph ANC Error: %s", err, extra={'user': 'SGT Logs'})
        return None


class GraphAnalyzer(ProgressUpdate):
    """
    A class that computes all the user-selected graph theory metrics and writes the results in a PDF file.

    Args:
        :param imp: Image Processor object.
        allow_multiprocessing: a decision to allow multiprocessing computing.
    """

    def __init__(self, imp: ImageProcessor, allow_multiprocessing: bool = True, use_igraph: bool = True):
        """
        A class that computes all the user-selected graph theory metrics and writes the results in a PDF file.

        :param imp: Image Processor object.
        :param allow_multiprocessing: Allows multiprocessing computing.
        :param use_igraph: Whether to use igraph C library module.

        >>> i_path = "path/to/image"
        >>> cfg_file = "path/to/sgt_configs.ini"
        >>>
        >>> def print_update(progress_val, progress_msg):
        >>>     print(f"{progress_val}: {progress_msg}")
        >>>
        >>> ntwk_obj, _ = ImageProcessor.from_image_file(i_path)
        >>> metrics_obj = GraphAnalyzer(ntwk_obj)
        >>> GraphAnalyzer.safe_run_analyzer(metrics_obj, print_update, save_to_pdf=True)

        """
        super(GraphAnalyzer, self).__init__()
        self._configs: dict = load_gtc_configs(imp.config_file)  # graph theory computation parameters and options.
        self._props: list = []
        self._allow_mp: bool = allow_multiprocessing
        self._use_igraph: bool = use_igraph
        self._ntwk_p: ImageProcessor = imp
        self._plot_figures: list | None = None
        self._results_df: None | pd.DataFrame = None
        self._weighted_results_df: None | pd.DataFrame = None
        self._scaling_results: dict = {}
        self._histogram_data = {"degree_distribution": [0], "clustering_coefficients": [0],
                               "betweenness_distribution": [0], "closeness_distribution": [0],
                               "eigenvector_distribution": [0], "ohms_distribution": [0],
                               "percolation_distribution": [], "weighted_degree_distribution": [0],
                               "weighted_clustering_coefficients": [0], "weighted_betweenness_distribution": [0],
                               "currentflow_distribution": [0], "weighted_closeness_distribution": [0],
                               "weighted_eigenvector_distribution": [0], "weighted_percolation_distribution": [0]}

    @property
    def configs(self) -> dict:
        """Returns the dictionary containing the parameters and options for computing graph theory metrics."""
        return self._configs

    @configs.setter
    def configs(self, configs: dict):
        """Sets the parameters and options for computing graph theory metrics."""
        self._configs = configs

    @property
    def props(self) -> list:
        """Returns the list of properties computed by StructuralGT."""
        return self._props

    @property
    def ntwk_p(self) -> ImageProcessor:
        """Returns the ImageProcessor object."""
        return self._ntwk_p

    @property
    def plot_figures(self) -> list:
        """Returns the list of figures generated by the graph theory metrics computation."""
        return self._plot_figures

    @property
    def results_df(self) -> pd.DataFrame:
        """Returns the Pandas DataFrame containing the results of the graph theory metrics computation."""
        return self._results_df

    @property
    def weighted_results_df(self) -> pd.DataFrame:
        """Returns the Pandas DataFrame containing the weighted results of the graph theory metrics computation."""
        return self._weighted_results_df

    @property
    def scaling_results(self) -> dict:
        return self._scaling_results

    def track_img_progress(self, value, msg) -> None:
        self.update_status([value, msg])

    def run_analyzer(self) -> None:
        """
            Execute functions that will process image filters and extract the graph from the processed image
        """

        # 1. Get graph extracted from selected images
        graph_obj = self._ntwk_p.graph_obj

        # 2. Apply image filters and extract the graph (only if it has not been executed)
        if graph_obj.nx_giant_graph is None:
            self._ntwk_p.add_listener(self.track_img_progress)
            self._ntwk_p.apply_img_filters()  # Apply image filters
            self._ntwk_p.build_graph_network()  # Extract graph from binary image
            self._ntwk_p.remove_listener(self.track_img_progress)
            self.abort = self._ntwk_p.abort
            self.update_status([100, "Graph successfully extracted!"]) if not self.abort else None
            graph_obj = self._ntwk_p.graph_obj

        if self.abort:
            return

        # 3a. Compute Unweighted GT parameters
        self._results_df = self.compute_gt_metrics(graph_obj.nx_giant_graph)  # replace with graph_obj.nx_giant_graph

        # 3b. Compute Scaling Scatter Plots
        scaling_data = None
        if self._configs["compute_scaling_behavior"]["value"] == 1:
            scaling_data = self.compute_scaling_data(full_img_df=self._results_df.copy())

        if self.abort:
            self.update_status([-1, "Problem encountered while computing un-weighted GT parameters."])
            return

        # 4. Compute Weighted GT parameters (skip if MultiGraph)
        self._weighted_results_df = self.compute_weighted_gt_metrics(graph_obj)

        if self.abort:
            self.update_status([-1, "Problem encountered while computing weighted GT parameters."])
            return

        # 5. Generate results in PDF
        self._plot_figures = self.generate_pdf_output(graph_obj, scaling_data)

        # 6. Save GT compute metrics into props
        self.get_compute_props()

    def compute_gt_metrics(self, graph: nx.Graph = None, save_histogram: bool = True, silent: bool = False) -> None|pd.DataFrame:
        """
        Compute unweighted graph theory metrics.

        :param graph: NetworkX graph object.
        :param save_histogram: Whether to save the histogram data.
        :param silent: Whether to send progress status and message (or silence them).

        :return: A Pandas DataFrame containing the unweighted graph theory metrics.
        """

        if graph is None:
            return None

        self.update_status([1, "Performing un-weighted GT parameters..."]) if not silent else None

        opt_gtc = self._configs
        data_dict = {"parameter": [], "value": []}

        node_count = int(nx.number_of_nodes(graph))
        edge_count = int(nx.number_of_edges(graph))

        data_dict["parameter"].append("Number of nodes")
        data_dict["value"].append(node_count)

        data_dict["parameter"].append("Number of edges")
        data_dict["value"].append(edge_count)

        """
        # length of edges
        length_arr = np.array(list(nx.get_edge_attributes(graph, 'length').values()))
        data_dict["parameter"].append('Average length (nm)')
        data_dict["value"].append(round(np.average(length_arr), 3))
        data_dict["parameter"].append('Median length (nm)')
        data_dict["value"].append(round(np.median(length_arr), 3))

        # width of edges
        width_arr = np.array(list(nx.get_edge_attributes(graph, 'width').values()))
        data_dict["parameter"].append('Average width (nm)')
        data_dict["value"].append(round(np.average(width_arr), 3))
        data_dict["parameter"].append('Median width (nm)')
        data_dict["value"].append(round(np.median(width_arr), 3))
        """

        # angle of edges (inbound and outbound)
        angle_arr = np.array(list(nx.get_edge_attributes(graph, 'angle').values()))
        data_dict["parameter"].append('Average edge angle (degrees)')
        data_dict["value"].append(round(np.average(angle_arr), 3))
        data_dict["parameter"].append('Median edge angle (degrees)')
        data_dict["value"].append(round(np.median(angle_arr), 3))

        if graph.number_of_nodes() <= 0:
            self.update_status([-1, "Problem with graph (change filter and graph options)."]) if not silent else None
            return None

        # creating degree histogram
        if opt_gtc["display_degree_histogram"]["value"] == 1:
            self.update_status([5, "Computing graph degree..."]) if not silent else None
            deg_distribution_1 = dict(nx.degree(graph))
            deg_distribution = np.array(list(deg_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["degree_distribution"] = deg_distribution
            data_dict["parameter"].append("Average degree")
            data_dict["value"].append(round(np.average(deg_distribution), 5))

        is_connected = None
        if (opt_gtc["compute_network_diameter"]["value"] == 1) or (
                opt_gtc["compute_avg_node_connectivity"]["value"] == 1):
            try:
                is_connected = nx.is_connected(graph)
            except nx.exception.NetworkXPointlessConcept:
                pass

        # calculating network diameter
        if opt_gtc["compute_network_diameter"]["value"] == 1:
            self.update_status([10, "Computing network diameter..."]) if not silent else None
            if is_connected:
                dia = int(diameter(graph))
            else:
                dia = np.nan
            data_dict["parameter"].append("Network diameter")
            data_dict["value"].append(dia)

        # calculating average nodal connectivity
        if opt_gtc["compute_avg_node_connectivity"]["value"] == 1:
            if self.abort:
                self.update_status([-1, "Task aborted."]) if not silent else None
                return None
            self.update_status([15, "Computing node connectivity..."]) if not silent else None
            avg_node_con = self.compute_avg_node_connectivity(graph, is_connected)
            data_dict["parameter"].append("Average node connectivity")
            data_dict["value"].append(avg_node_con)

        # calculating graph density
        if opt_gtc["compute_graph_density"]["value"] == 1:
            self.update_status([20, "Computing graph density..."]) if not silent else None
            g_density = nx.density(graph)
            g_density = round(g_density, 5)
            data_dict["parameter"].append("Graph density")
            data_dict["value"].append(g_density)

        # calculating global efficiency
        if opt_gtc["compute_global_efficiency"]["value"] == 1:
            if self.abort:
                self.update_status([-1, "Task aborted."])
                return None
            self.update_status([25, "Computing global efficiency..."]) if not silent else None
            g_eff = global_efficiency(graph)
            g_eff = round(g_eff, 5)
            data_dict["parameter"].append("Global efficiency")
            data_dict["value"].append(g_eff)

        if opt_gtc["compute_wiener_index"]["value"] == 1:
            self.update_status([30, "Computing wiener index..."]) if not silent else None
            # settings.update_label("Calculating w_index...")
            w_index = wiener_index(graph)
            w_index = round(w_index, 1)
            data_dict["parameter"].append("Wiener Index")
            data_dict["value"].append(w_index)

        # calculating assortativity coefficient
        if opt_gtc["compute_assortativity_coef"]["value"] == 1:
            self.update_status([35, "Computing assortativity coefficient..."]) if not silent else None
            a_coef = degree_assortativity_coefficient(graph)
            a_coef = round(a_coef, 5)
            data_dict["parameter"].append("Assortativity coefficient")
            data_dict["value"].append(a_coef)

        # calculating clustering coefficients
        if opt_gtc["compute_avg_clustering_coef"]["value"] == 1:
            self.update_status([40, "Computing clustering coefficients..."]) if not silent else None
            coefficients_1 = clustering(graph)
            cl_coefficients = np.array(list(coefficients_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["clustering_coefficients"] = cl_coefficients
            data_dict["parameter"].append("Average clustering coefficient")
            data_dict["value"].append(round(np.average(cl_coefficients), 5))

        # calculating betweenness centrality histogram
        if opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1:
            self.update_status([45, "Computing betweenness centrality..."]) if not silent else None
            b_distribution_1 = betweenness_centrality(graph)
            b_distribution = np.array(list(b_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["betweenness_distribution"] = b_distribution
            data_dict["parameter"].append("Average betweenness centrality")
            data_dict["value"].append(round(np.average(b_distribution), 5))

        # calculating eigenvector centrality
        if opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1:
            self.update_status([50, "Computing eigenvector centrality..."]) if not silent else None
            try:
                e_vecs_1 = eigenvector_centrality(graph, max_iter=100)
            except nx.exception.PowerIterationFailedConvergence:
                e_vecs_1 = eigenvector_centrality(graph, max_iter=10000)
            e_vecs = np.array(list(e_vecs_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["eigenvector_distribution"] = e_vecs
            data_dict["parameter"].append("Average eigenvector centrality")
            data_dict["value"].append(round(np.average(e_vecs), 5))

        # calculating closeness centrality
        if opt_gtc["display_closeness_centrality_histogram"]["value"] == 1:
            self.update_status([55, "Computing closeness centrality..."]) if not silent else None
            close_distribution_1 = closeness_centrality(graph)
            close_distribution = np.array(list(close_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["closeness_distribution"] = close_distribution
            data_dict["parameter"].append("Average closeness centrality")
            data_dict["value"].append(round(np.average(close_distribution), 5))

        # calculating Ohms centrality
        if opt_gtc["display_ohms_histogram"]["value"] == 1:
            self.update_status([60, "Computing Ohms centrality..."]) if not silent else None
            o_distribution_1, res = self.compute_ohms_centrality(graph)
            o_distribution = np.array(list(o_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["ohms_distribution"] = o_distribution
            data_dict["parameter"].append("Average Ohms centrality")
            data_dict["value"].append(round(np.average(o_distribution), 5))
            data_dict["parameter"].append("Ohms centrality -- avg. area " + r"($m^2$)")
            data_dict["value"].append(round(res['avg area'], 5))
            data_dict["parameter"].append("Ohms centrality -- avg. length (m)")
            data_dict["value"].append(round(res['avg length'], 5))
            data_dict["parameter"].append("Ohms centrality -- avg. width (m)")
            data_dict["value"].append(round(res['avg width'], 5))
            data_dict["parameter"].append("Ohms centrality -- g shape coeff.")
            data_dict["value"].append(round(res['g shape'], 5))
            data_dict["parameter"].append("Ohms centrality -- conductivity (S/m)")
            data_dict["value"].append(round(res['conductivity'], 5))

        return pd.DataFrame(data_dict)

    def compute_weighted_gt_metrics(self, graph_obj: FiberNetworkBuilder = None, save_histogram: bool = True, silent: bool = False) -> None|pd.DataFrame:
        """
        Compute weighted graph theory metrics.

        :param graph_obj: GraphExtractor object.
        :param save_histogram: Whether to save histogram data.
        :param silent: Whether to send progress status and message (or silence them).

        :return: A Pandas DataFrame containing the weighted graph theory metrics.
        """
        if graph_obj is None:
            return None

        if not graph_obj.configs["has_weights"]["value"]:
            return None

        self.update_status([70, "Performing weighted GT parameters..."]) if not silent else None

        graph = graph_obj.nx_giant_graph
        opt_gtc = self._configs
        wt_type = graph_obj.get_weight_type()
        weight_type = FiberNetworkBuilder.get_weight_options().get(wt_type)
        data_dict = {"parameter": [], "value": []}

        if graph.number_of_nodes() <= 0:
            self.update_status([-1, "Problem with graph (change filter and graph options)."]) if not silent else None
            return None

        if opt_gtc["display_degree_histogram"]["value"] == 1:
            self.update_status([72, "Compute weighted graph degree..."]) if not silent else None
            deg_distribution_1 = dict(nx.degree(graph, weight='weight'))
            deg_distribution = np.array(list(deg_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["weighted_degree_distribution"] = deg_distribution
            data_dict["parameter"].append(f"{weight_type}-weighted average degree")
            data_dict["value"].append(round(np.average(deg_distribution), 5))

        if opt_gtc["compute_wiener_index"]["value"] == 1:
            self.update_status([74, "Compute weighted wiener index..."]) if not silent else None
            w_index = wiener_index(graph, weight='length')
            w_index = round(w_index, 1)
            data_dict["parameter"].append("Length-weighted Wiener Index")
            data_dict["value"].append(w_index)

        if opt_gtc["compute_avg_node_connectivity"]["value"] == 1:
            self.update_status([76, "Compute weighted node connectivity..."]) if not silent else None
            connected_graph = nx.is_connected(graph)
            if connected_graph:
                max_flow = float(0)
                p = periphery(graph)
                q = len(p) - 1
                for s in range(0, q - 1):
                    for t in range(s + 1, q):
                        flow_value = maximum_flow(graph, p[s], p[t], capacity='weight')[0]
                        if flow_value > max_flow:
                            max_flow = flow_value
                max_flow = round(max_flow, 5)
            else:
                max_flow = np.nan
            data_dict["parameter"].append("Max flow between periphery")
            data_dict["value"].append(max_flow)

        if opt_gtc["compute_assortativity_coef"]["value"] == 1:
            self.update_status([78, "Compute weighted assortativity..."]) if not silent else None
            a_coef = degree_assortativity_coefficient(graph, weight='width')
            a_coef = round(a_coef, 5)
            data_dict["parameter"].append("Width-weighted assortativity coefficient")
            data_dict["value"].append(a_coef)

        if opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1:
            self.update_status([80, "Compute weighted betweenness centrality..."]) if not silent else None
            b_distribution_1 = betweenness_centrality(graph, weight='weight')
            b_distribution = np.array(list(b_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["weighted_betweenness_distribution"] = b_distribution
            data_dict["parameter"].append(f"{weight_type}-weighted betweenness centrality")
            data_dict["value"].append(round(np.average(b_distribution), 5))

        if opt_gtc["display_closeness_centrality_histogram"]["value"] == 1:
            self.update_status([82, "Compute weighted closeness centrality..."]) if not silent else None
            close_distribution_1 = closeness_centrality(graph, distance='length')
            close_distribution = np.array(list(close_distribution_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["weighted_closeness_distribution"] = close_distribution
            data_dict["parameter"].append(f"Length-weighted average closeness centrality")
            data_dict["value"].append(round(np.average(close_distribution), 5))

        if opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1:
            if self.abort:
                self.update_status([-1, "Task aborted."]) if not silent else None
                return None
            self.update_status([84, "Compute weighted eigenvector centrality..."]) if not silent else None
            try:
                e_vecs_1 = eigenvector_centrality(graph, max_iter=100, weight='weight')
            except nx.exception.PowerIterationFailedConvergence:
                e_vecs_1 = eigenvector_centrality(graph, max_iter=10000, weight='weight')
            e_vecs = np.array(list(e_vecs_1.values()), dtype=float)
            if save_histogram:
                self._histogram_data["weighted_eigenvector_distribution"] = e_vecs
            data_dict["parameter"].append(f"{weight_type}-weighted average eigenvector centrality")
            data_dict["value"].append(round(np.average(e_vecs), 5))

        # calculate cross-sectional area of edges
        wt_type = graph_obj.get_weight_type()
        if wt_type == 'AREA':
            self.update_status([88, "Computing average (edge) cross-sectional area..."]) if not silent else None
            temp_distribution = []
            for (s, e) in graph.edges():
                temp_distribution.append(graph[s][e]['weight'])
            a_distribution = np.array(temp_distribution, dtype=float)
            ae_val = np.average(a_distribution)
            ae_val = round(ae_val, 5)
            data_dict["parameter"].append(f"Average edge cross-sectional area (nm\u00b2)")
            data_dict["value"].append(ae_val)

        return pd.DataFrame(data_dict)

    def compute_scaling_data(self, full_img_df: pd.DataFrame = None) -> defaultdict:
        """"""
        self.update_status([65, "Computing scaling behaviour..."])
        self._ntwk_p.add_listener(self.track_img_progress)
        num_filters = int(self._configs["scaling_behavior_kernel_count"]["value"])
        num_patches = int(self._configs["scaling_behavior_patches_per_kernel"]["value"])
        calc_avg = self._configs["scaling_behavior_compute_avg"]["value"]
        graph_groups = self._ntwk_p.build_graph_from_patches(num_kernels=num_filters, patch_count_per_kernel=num_patches, compute_avg=calc_avg)
        self._ntwk_p.remove_listener(self.track_img_progress)

        sorted_plt_data = defaultdict(lambda: defaultdict(list))
        avg_df = None
        for (h, w), nx_graphs in graph_groups.items():
            num_graphs = len(nx_graphs)
            for i, nx_graph in enumerate(nx_graphs):
                self.update_status([101, f"Computing GT parameters for filter {h}x{w}: graph-patch {i + 1}/{num_graphs}..."])
                temp_df = self.compute_gt_metrics(nx_graph, save_histogram=False, silent=True)
                if temp_df is None:
                    # Skip the problematic graph
                    continue
                for _, row in temp_df.iterrows():
                    x_param = row["parameter"]
                    y_value = row["value"]
                    if ' edge angle' in x_param:  # Skip this
                        continue
                    sorted_plt_data[x_param][h].append(y_value) if num_graphs > 4 else None

                # Save GT parameters/descriptors of 90% image to DF
                if num_graphs > 4:
                    continue
                else:
                    temp_df = temp_df.rename(columns={'value': f'value-{i + 1}'})
                    if i == 0:
                        avg_df = temp_df
                    else:
                        avg_df = avg_df.merge(temp_df, on='parameter')
        
        # Include the computed GT descriptors of the entire image
        if full_img_df is not None:
            # Get full image dimensions
            h, w = self._ntwk_p.binary_image_2d.shape
            for _ in range(num_patches):
                for _, row in full_img_df.iterrows():
                    x_param = row["parameter"]
                    y_value = row["value"]
                    # print(f"{x_param}-{h}: {y_value}")
                    sorted_plt_data[x_param][h].append(y_value)

        # Add average to scaling results (for the Excel file)
        if avg_df is not None:
            self._scaling_results["SGT Descriptors"] = avg_df

        return sorted_plt_data

    def compute_ohms_centrality(self, nx_graph: nx.Graph) -> tuple[dict, dict] | tuple[None, None]:
        r"""
        Computes Ohms centrality value for each node based on actual pixel width and length of edges in meters.

        :param nx_graph: NetworkX graph object.

        Returns: Ohms centrality distribution
        """
        ohms_dict = {}
        lst_area = []
        lst_len = []
        lst_width = []

        if self._ntwk_p.selected_batch.is_graph_only:
            return None, None
        sel_images = self._ntwk_p.selected_images
        px_sizes = np.array([img.configs["pixel_width"]["value"] for img in sel_images])
        rho_dims = np.array([img.configs["resistivity"]["value"] for img in sel_images])

        px_size = float(np.average(px_sizes.astype(float)))
        rho_dim = float(np.average(rho_dims.astype(float)))
        pixel_dim = px_size  # * (10 ** 9)  # Convert to nanometers
        g_shape = 1

        b_dict = betweenness_centrality(nx_graph)
        lst_nodes = list(nx_graph.nodes())
        for n in lst_nodes:
            # compute Ohms centrality value for each node
            b_val = float(b_dict[n])
            if b_val == 0:
                ohms_val = 0
            else:
                connected_nodes = nx_graph[n]  # all nodes connected to node n
                arr_len = []
                arr_dia = []
                for idx, val in connected_nodes.items():
                    arr_len.append(val['length'])
                    arr_dia.append(val['width'])
                arr_len = np.array(arr_len, dtype=float)
                arr_dia = np.array(arr_dia, dtype=float)

                pix_width = float(np.average(arr_dia))
                pix_length = np.sum(arr_len)
                length = pix_length * pixel_dim
                width = pix_width * pixel_dim
                # area = math.pi * 89.6 * (width * 0.5) ** 2
                area = g_shape * (width * width)
                ohms_val = ((b_val * length * rho_dim) / area)
                lst_len.append(length)
                lst_area.append(area)
                lst_width.append(width)
                # if n < 5:
            ohms_dict[n] = ohms_val
        avg_area = np.average(np.array(lst_area, dtype=float))
        med_area = np.median(np.array(lst_area, dtype=float))
        avg_len = np.average(np.array(lst_len, dtype=float))
        med_len = np.median(np.array(lst_len, dtype=float))
        avg_width = np.average(np.array(lst_width, dtype=float))
        med_width = np.median(np.array(lst_width, dtype=float))
        res = {
            'avg area': avg_area, 'med area': med_area,
            'avg length': avg_len, 'med length': med_len,
            'avg width': avg_width, 'med width': med_width,
            'g shape': g_shape, 'conductivity': (1 / rho_dim)}

        return ohms_dict, res

    def compute_avg_node_connectivity(self, nx_graph: nx.Graph, is_graph_connected=False) -> float:
        r"""Returns the average connectivity of a graph G.

        The average connectivity `\bar{\kappa}` of a graph G is the average
        of local node connectivity over all pairs of the nx_graph nodes.

        :param nx_graph: NetworkX graph object.
        :param is_graph_connected: Boolean
        """

        def nx_average_node_connectivity(flow_func=None):
            r"""Returns the average connectivity of a graph G.

            The average connectivity `\bar{\kappa}` of a graph G is the average
            of local node connectivity over all pairs of the nx_graph nodes.

            https://networkx.org/documentation/stable/_modules/networkx/algorithms/connectivity/connectivity.html#average_node_connectivity

            Parameters
            ----------
            :param flow_func : Function
                A function for computing the maximum flow between a pair of nodes.
                The function has to accept at least three parameters: a Digraph,
                a source node, and a target node. And return a residual network
                that follows NetworkX conventions (see: meth:`maximum_flow` for
                details). If flow_func is None, the default maximum flow function
                (: meth:`edmonds_karp`) is used. See :meth:`local_node_connectivity`
                for details. The choice of the default function may change from
                version to version and should not be relied on. Default value: None.

            Returns
            -------
            K : float
                Average node connectivity

            References
            ----------
            [1]  Beineke, L., O. Oellermann, and r_network. Pippert (2002). The average
                    connectivity of a graph. Discrete mathematics 252(1-3), 31-45.
                    https://www.sciencedirect.com/science/article/pii/S0012365X01001807

            """

            if nx_graph.is_directed():
                iter_func = itertools.permutations
            else:
                iter_func = itertools.combinations

            # Reuse the auxiliary digraph and the residual network
            a_digraph = nx.algorithms.connectivity.build_auxiliary_node_connectivity(nx_graph)
            r_network = nx.algorithms.flow.build_residual_network(a_digraph, "capacity")
            # kwargs = {"flow_func": flow_func, "auxiliary": a_digraph, "residual": r_network}

            total, count = 0, 0
            with multiprocessing.Pool() as pool:
                items = [(nx_graph, u, v, flow_func, a_digraph, r_network) for u, v in iter_func(nx_graph, 2)]
                async_result = pool.starmap_async(nx.algorithms.connectivity.local_node_connectivity, items)
                for n in async_result.get():
                    total += n
                    count += 1
                    if self.abort:
                        self.update_status([-1, "Task aborted."])
                        pool.terminate()
                        pool.join()
                        return 0
                    if n is not None:
                        total += n
                        count += 1
            anc = total / count if count > 0 else 0
            return anc

        def igraph_average_node_connectivity():
            r"""
            Returns the average connectivity of a graph G.

            The average connectivity of a graph G is the average
            of local node connectivity over all pairs of the Graph (G) nodes.
            """

            ig_graph = ig.Graph.from_networkx(nx_graph)
            num_nodes = ig_graph.vcount()
            total, count = 0, 0
            with multiprocessing.Pool() as pool:
                # Prepare all node pairs (i < j)
                items = [(ig_graph, i, j) for i in range(num_nodes) for j in range(i + 1, num_nodes)]
                async_result = pool.starmap_async(_worker_vertex_connectivity, items)
                for n in async_result.get():
                    if self.abort:
                        self.update_status([-1, "Task aborted."])
                        pool.terminate()
                        pool.join()
                        return 0
                    if n is not None:
                        total += n
                        count += 1
            anc = total / count if count > 0 else 0
            return anc

        def igraph_clang_average_node_connectivity():
            r"""Returns the average connectivity of a graph G.

            The average connectivity of a graph G is the average
            of local node connectivity over all pairs of the Graph (G) nodes.

            """
            from .c_lang import sgt_c_module as sgt

            cpu_count = get_num_cores()
            num_threads = cpu_count if nx.number_of_nodes(nx_graph) < 2000 else cpu_count * 2
            anc = 0

            try:
                filename, output_location = self._ntwk_p.get_filenames()
                g_filename = filename + "_graph.txt"
                graph_file = os.path.join(output_location, g_filename)
                nx.write_edgelist(nx_graph, graph_file, data=False)
                anc = sgt.compute_anc(graph_file, num_threads, self._allow_mp)
            except Exception as err:
                logging.exception("Computing ANC Error: %s", err, extra={'user': 'SGT Logs'})
            return anc

        if is_graph_connected:
            # use_igraph = opt_gtc["computing_lang == 'C'"]["value"]
            if self._use_igraph:
                # use iGraph Lib in C
                self.update_status([15, "Using iGraph library..."])
                try:
                    avg_node_con = igraph_clang_average_node_connectivity()
                except ImportError:
                    avg_node_con = igraph_average_node_connectivity()
            else:
                # Use NetworkX Lib in Python
                self.update_status([15, "Using NetworkX library..."])
                if self._allow_mp:  # Multi-processing
                    avg_node_con = nx_average_node_connectivity()
                else:
                    avg_node_con = average_node_connectivity(nx_graph)
            avg_node_con = round(avg_node_con, 5)
        else:
            avg_node_con = np.nan
        return avg_node_con

    def compute_graph_conductance(self, graph_obj):
        """
        Computes graph conductance through an approach based on eigenvectors or spectral frequency.
        Implements ideas proposed in:    https://doi.org/10.1016/j.procs.2013.09.311.

        Conductance can closely be approximated via eigenvalue computation,
        a fact which has been well-known and well-used in the graph theory community.

        The Laplacian matrix of a directed graph is by definition generally non-symmetric,
        while, e.g., traditional spectral clustering is primarily developed for undirected
        graphs with symmetric adjacency and Laplacian matrices. A trivial approach to applying the
        techniques requiring symmetry is to turn the original directed graph into an
        undirected graph and build the Laplacian matrix for the latter.

        We need to remove isolated nodes (to avoid singular adjacency matrix).
        The degree of a node is the number of edges incident to that node.
        When a node has a degree of zero, it means that there are no edges
        connected to that node. In other words, the node is isolated from
        the rest of the graph.

        :param graph_obj: Graph Extractor object.

        """
        self.update_status([101, "Computing graph conductance..."])
        # Make a copy of the graph
        graph = graph_obj.nx_giant_graph.copy()
        weighted = graph_obj.configs["has_weights"]["value"]

        # It is important to notice our graph is (mostly) a directed graph,
        # meaning that it is: (asymmetric) with self-looping nodes

        # 1. Remove self-looping edges from the graph, they cause zero values in Degree matrix.
        # 1a. Get Adjacency matrix
        adj_mat = nx.adjacency_matrix(graph).todense()

        # 1b. Remove (self-loops) non-zero diagonal values in Adjacency matrix
        np.fill_diagonal(adj_mat, 0)

        # 1c. Create the new graph
        giant_graph = nx.from_numpy_array(adj_mat)

        # 2a. Identify isolated nodes
        isolated_nodes = list(nx.isolates(giant_graph))

        # 2b. Remove isolated nodes
        giant_graph.remove_nodes_from(isolated_nodes)

        # 3a. Check the connectivity of the graph
        # It has less than two nodes or is not connected.
        # Identify connected components
        connected_components = list(nx.connected_components(graph))
        if not connected_components:  # In case the graph is empty
            connected_components = []
        sub_graphs = [graph.subgraph(c).copy() for c in connected_components]

        giant_graph = max(sub_graphs, key=lambda g: g.number_of_nodes())

        # 4. Compute normalized-laplacian matrix
        if weighted:
            norm_laplacian_matrix = nx.normalized_laplacian_matrix(giant_graph, weight='weight').toarray()
        else:
            # norm_laplacian_matrix = compute_norm_laplacian_matrix(giant_graph)
            norm_laplacian_matrix = nx.normalized_laplacian_matrix(giant_graph).toarray()

        # 5. Compute eigenvalues
        # e_vals, _ = xp.linalg.eig(norm_laplacian_matrix)
        e_vals = sp.linalg.eigvals(norm_laplacian_matrix)

        # 6. Approximate conductance using the 2nd smallest eigenvalue
        # 6a. Compute the minimum and maximum values of graph conductance.
        sorted_vals = np.array(e_vals.real)
        sorted_vals.sort()
        # approximate conductance using the 2nd smallest eigenvalue
        try:
            # Maximum Conductance
            val_max = math.sqrt((2 * sorted_vals[1]))
        except ValueError:
            val_max = np.nan
        # Minimum Graph Conductance
        val_min = sorted_vals[1] / 2

        return val_max, val_min

    def get_compute_props(self) -> None:
        """
        A method that retrieves graph theory computed parameters and stores them in a list-array.

        Returns: list of computed GT params.

        """
        self._props = []
        # 1. Unweighted parameters
        if self._results_df is None:
            return
        param_df = self._results_df.copy()
        self._props.append(['UN-WEIGHTED', 'PARAMETERS'])
        for _, row in param_df.iterrows():
            x_param = row["parameter"]
            y_value = row["value"]
            self._props.append([x_param, y_value])

        # 2. Weighted parameters
        if self._weighted_results_df is None:
            return
        param_df = self._weighted_results_df.copy()
        self._props.append(['WEIGHTED', 'PARAMETERS'])
        for _, row in param_df.iterrows():
            x_param = row["parameter"]
            y_value = row["value"]
            self._props.append([x_param, y_value])

    def generate_pdf_output(self, graph_obj: FiberNetworkBuilder, scaling_data=None) -> list[plt.Figure]:
        """
        Generate results as graphs and plots which should be written in a PDF file.

        :param graph_obj: Graph extractor object.
        :param scaling_data: Computed scaling results stored in a dictionary.

        :return: List of results.
        """

        self.update_status([90, "Generating PDF GT Output..."])
        opt_gtc = self._configs
        out_figs = []

        sel_images = self._ntwk_p.selected_images
        img_3d = np.asarray(self._ntwk_p.image_3d) if len(sel_images) > 0 else None

        def plot_gt_results():
            """
            Create a table of weighted and unweighted graph theory results.

            :return: Matplotlib figures of unweighted and weighted graph theory results.
            """

            opt_gte = graph_obj.configs
            data = self._results_df
            w_data = self._weighted_results_df

            plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
            ax = plt_fig.add_subplot(1, 1, 1)
            ax.set_axis_off()
            ax.set_title("Unweighted GT parameters")
            col_width = [2 / 3, 1 / 3]
            tab_1 = tbl.table(ax, cellText=data.values[:, :], loc='upper center', colWidths=col_width, cellLoc='left')
            tab_1.scale(1, 1.5)

            if opt_gte["has_weights"]["value"] == 1 and w_data is not None:
                plt_fig_wt = plt.Figure(figsize=(8.5, 11), dpi=300)
                ax = plt_fig_wt.add_subplot(1, 1, 1)
                ax.set_axis_off()
                ax.set_title("Weighted GT parameters")
                tab_2 = tbl.table(ax, cellText=w_data.values[:, :], loc='upper center', colWidths=col_width,
                                  cellLoc='left')
                tab_2.scale(1, 1.5)
            else:
                plt_fig_wt = None
            return plt_fig, plt_fig_wt

        def plot_bin_images():
            """
            Create plot figures of original, processed, and binary image.

            :return:
            """

            plt_figs = []
            is_3d = True if len(sel_images) > 1 else False

            for i, img in enumerate(sel_images):
                raw_img = img.img_2d
                filtered_img = img.img_mod
                img_bin = img.img_bin

                plt_fig = plt.Figure(figsize=(8.5, 8.5), dpi=400)
                ax_1 = plt_fig.add_subplot(2, 2, 1)
                ax_2 = plt_fig.add_subplot(2, 2, 2)
                ax_3 = plt_fig.add_subplot(2, 2, 3)
                ax_4 = plt_fig.add_subplot(2, 2, 4)

                ax_1.set_title(f"Frame {i}: Original Image") if is_3d else ax_1.set_title(f"Original Image")
                ax_1.set_axis_off()
                ax_1.imshow(raw_img, cmap='gray')

                ax_2.set_title(f"Frame {i}: Processed Image") if is_3d else ax_2.set_title(f"Processed Image")
                ax_2.set_axis_off()
                ax_2.imshow(filtered_img, cmap='gray')

                ax_3.set_title(f"Frame {i}: Binary Image") if is_3d else ax_3.set_title(f"Binary Image")
                ax_3.set_axis_off()
                ax_3.imshow(img_bin, cmap='gray')

                img.plot_img_histogram(axes=ax_4)
                hist_title = f"Frame {i}: Histogram of Processed Image" if is_3d else f"Histogram of Processed Image"
                ax_4.set_title(hist_title)

                plt_figs.append(plt_fig)
            return plt_figs

        def plot_run_configs():
            """
            Create a page (as a figure) that will show the user-selected parameters and options.

            :return: A Matplotlib figure object.
            """

            plt_fig = plt.Figure(figsize=(8.5, 8.5), dpi=300)
            ax = plt_fig.add_subplot(1, 1, 1)
            ax.set_axis_off()
            ax.set_title("Run Info")

            # similar to the start of the csv file, this is just getting all the relevant settings to display in the PDF
            _, filename = os.path.split(self._ntwk_p.img_path)
            now = datetime.datetime.now()

            run_info = ""
            run_info += filename + "\n"
            run_info += now.strftime("%Y-%m-%d %H:%M:%S") + "\n----------------------------\n\n"

            # Image Configs
            # sel_img_batch = self.ntwk_p.selected_batch
            run_info += self._ntwk_p.image_obj.get_config_info()  # Get configs of first image
            run_info += "\n\n"

            # Graph Configs
            run_info += graph_obj.get_config_info()
            run_info += "\n\n"

            ax.text(0.5, 0.5, run_info, horizontalalignment='center', verticalalignment='center')
            return plt_fig

        def plot_scaling_behavior():
            """"""
            self.update_status([91, "Plotting scaling behavior..."])

            # Define our 'best-fit' model
            def power_law_model(x, a, k):
                """
                    A best-fit model that follows the power law distribution: y = a * x^(-k),
                    where a and k are fitting parameters.

                    Args:
                        x (np.array): Array of x values
                        a (float): intercept or scale factor/constant (shifts curve up/down). It does not affect the shape/decay rate of the curve (only how high the curve starts)
                        k (float): exponent of the power law distribution. It defines how fast y decays as x increases.
                """
                return a * (x ** (-k))

            def truncated_power_law_model(x, a, k, c):
                """
                A best-fit model that follows the truncated power law distribution: y = a * x^(-k) * exp(-c * x),
                where a, c, and k are fitting parameters.

                https://en.wikipedia.org/wiki/Power_law#Power_law_with_exponential_cutoff

                Args:
                    x (np.array): Array of x values
                    a (float): fitting parameter
                    k (float): fitting parameter
                    c (float): cut-off fitting parameter
                """
                return a * (x ** (-k)) * np.exp(-c * x)

            def lognormal_model(x, mu, sigma, a):
                """
                Log-normal model (Y depends on X, X is log-normal).

                Args:
                    x (np.array): Array of x values
                    mu (float): fitting parameter
                    sigma (float): fitting parameter
                    a (float): fitting parameter

                Returns:

                """
                return a * (1 / (x * sigma * np.sqrt(2 * np.pi))) * np.exp(-((np.log(x) - mu) ** 2) / (2 * sigma ** 2))

            def find_elbow(x, y):
                """"""

                try:
                    from kneed import KneeLocator

                    # First and second derivative
                    dy = np.gradient(y, x)
                    ddy = np.gradient(dy, x)

                    is_increasing = np.all(dy > 0)
                    is_decreasing = np.all(dy < 0)
                    is_convex = np.all(ddy > 0)
                    is_concave = np.all(ddy < 0)

                    if is_increasing:
                        direction = 'increasing'
                    elif is_decreasing:
                        direction = 'decreasing'
                    else:
                        direction = None

                    if is_convex:
                        curve = 'convex'
                    elif is_concave:
                        curve = 'concave'
                    else:
                        curve = None

                    # print(f"Curve: {curve}, Direction: {direction}")
                    if direction is None or curve is None:
                        return None
                    elbow = KneeLocator(x, y, S=1.0, curve=curve, direction=direction)
                    return elbow.knee
                except Exception as error:
                    logging.exception("Scaling Law (Scale Estimation) Error: %s", error, extra={'user': 'SGT Logs'})
                    return None

            def plot_axis(subplot_num, plt_type="", plot_err=True):
                """"""
                subplot_num += 1
                axis = plt_fig.add_subplot(2, 2, subplot_num)
                if plot_err:
                    axis.errorbar(x_avg, y_avg, xerr=x_err, yerr=y_err, label='Data', color='b', capsize=4, marker='s',
                                  markersize=4, linewidth=1, linestyle='-')
                axis.set_title(f"{plt_type}\nNodes vs {y_title}", fontsize=10)
                axis.set(xlabel='No. of Nodes', ylabel=f'{param_name}')
                # axis.legend()
                return axis, subplot_num

            # Initialize plot figures
            plt_figs = []
            if scaling_data is None:
                return plt_figs

            # Plot scaling behavior
            i = 0
            x_label, best_scale = None, None
            y_title = ""
            x_values, x_avg, x_err, x_fit = np.nan, np.nan, np.nan, np.nan
            data_df, fit_data_df = None, None
            plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
            for param_name, plt_dict in scaling_data.items():
                # axis title name
                y_title = param_name.split('(')[0] if '(' in param_name else param_name

                # Retrieve plot data
                kernel_dims = np.array(sorted(plt_dict.keys()))  # Optional: sort heights and save as a numpy array
                y_lst = [plt_dict[d] for d in kernel_dims]  # shape: (n_samples, n_kernels)

                # Pad with NaN
                max_len = max(len(row) for row in y_lst)
                padded_lst = [row + [np.nan] * (max_len - len(row)) for row in y_lst]

                # Convert to a Numpy array
                y_values = np.array(padded_lst).T
                y_avg = np.nanmean(y_values, axis=0)
                y_err = np.nanstd(y_values, axis=0, ddof=1) / np.sqrt(y_values.shape[0])
                if np.any(np.isnan(y_avg)):
                    # print(f"{param_name} has NaN values: {y_avg}")
                    continue

                # Plot (taking Node-count as the independent variable) against other parameters
                add_plot = False
                if x_label is None:
                    # First Param becomes X-axis: 'Number of Nodes'
                    x_label = param_name
                    # x_values = y_values
                    x_avg = y_avg
                    x_err = y_err
                    x_fit = np.linspace(min(x_avg), max(x_avg), 100)

                    # Estimate best scale
                    best_scale = find_elbow(kernel_dims[:-1], y_avg[:-1])

                    # Add plots to Figure
                    try:
                        # a) Plot Nodes vs. Kernel-size (scatter plot)
                        # Plot for each Parameter?
                        ax, i = plot_axis(i, "", plot_err=False)
                        ax.set_title(f"Kernel Size vs {y_title}", fontsize=10)
                        ax.set(xlabel='Kernel Size', ylabel=y_title)
                        ax.errorbar(kernel_dims, y_avg, yerr=y_err, label='Data', color='b', capsize=4, marker='s',
                                    markersize=4, linewidth=1, linestyle='-')

                        # b) Log-log plot of Kernel-size vs. Node-count
                        log_x = np.log10(kernel_dims)
                        log_y = np.log10(y_avg)
                        slope, intercept, r_value, p_value, std_err = sp.stats.linregress(log_x, log_y)
                        log_y_fit = slope * log_x + intercept  # Compute line of best-fit

                        ax = plt_fig.add_subplot(2, 2, i+1)
                        ax.set_title(f"Log-Log Plot of\nKernel Size vs {y_title}", fontsize=10)
                        ax.set(xlabel='Kernel Size', ylabel=y_title)
                        ax.plot(log_x, log_y, label='Data', color='b', marker='s', markersize=3)
                        ax.plot(log_x, log_y_fit, label=f'Fit: slope={slope:.2f}, $R^2$={r_value ** 2:.3f}', color='r')
                        ax.legend()

                        plt_figs.append(plt_fig)
                    except Exception as e:
                        logging.exception("Scaling Law (Kernel-Nodes) Error: %s", e, extra={'user': 'SGT Logs'})

                    # Write to DataFrame
                    data_df = pd.DataFrame({'kernel-dim': kernel_dims, 'x-avg': x_avg, 'x-std': x_err})
                    fit_data_df = pd.DataFrame({'x-fit': x_fit})
                else:
                    # 2a. Plot on the Log-Log scale
                    try:
                        # 1. Transform to log-log scale
                        log_x = np.log10(x_avg)
                        log_y = np.log10(y_avg)

                        # 2. Perform linear regression in log-log scale
                        slope, intercept, r_value, p_value, std_err = sp.stats.linregress(log_x, log_y)
                        log_y_fit = slope * log_x + intercept  # Compute line of best-fit

                        # 3. Plot data (Log-Log scale with the line best-fit)
                        ax, i = plot_axis(i, "Log-Log Plot of", plot_err=False)
                        ax.plot(log_x, log_y, label='Data', color='b', marker='s', markersize=3)
                        ax.plot(log_x, log_y_fit, label=f'Fit: slope={slope:.2f}, $R^2$={r_value ** 2:.3f}', color='r')
                        ax.legend()
                        add_plot = True

                        # Write to DataFrame
                        if data_df is not None:
                            data_df['y-avg'] = y_avg
                            data_df['y-std'] = y_err
                            data_df['log-x'] = log_x
                            data_df['log-y'] = log_y
                            data_df['log-y-fit'] = log_y_fit
                    except Exception as err:
                        add_plot = False
                        logging.exception("Scaling Law (Log-Log Fit) Error: %s", err, extra={'user': 'SGT Logs'})

                    if opt_gtc["scaling_behavior_power_law_fit"]["value"] == 1:
                        # 2b. Compute the line of best-fit on our data according to our power-law model
                        try:
                            init_params = [1.0, 1.0]  # initial guess for [a, k]
                            optimal_params: np.ndarray = \
                            sp.optimize.curve_fit(power_law_model, x_avg, y_avg, p0=init_params)[0]
                            a_fit, k_fit = float(optimal_params[0]), float(optimal_params[1])
                            # print(f"Fitted parameters: a = {a_fit:.4f}, k = {k_fit:.4f}")

                            # Generate points for the best-fit curve
                            y_fit_pwr = power_law_model(x_fit, a_fit, k_fit)

                            # Compute Kolmogorov-Smirnov Test & GoodnessOfFit P-Values
                            res_good_fit = sp.stats.goodness_of_fit(sp.stats.powerlaw, y_avg)
                            ks_stat, ks_p_val = res_good_fit.statistic, res_good_fit.pvalue

                            # 3b. Plot data (power-law best fit)
                            ax, i = plot_axis(i, "Power Law Fit and Plot of")
                            ax.plot(x_fit, y_fit_pwr,
                                    label=f'Fit: $y = ax^{{-k}}$\n$a={a_fit:.2f}, k={k_fit:.2f}$\nKS Stat={ks_stat:.2f}, P-Val={ks_p_val:.2f}',
                                    color='red')
                            ax.legend()

                            # Write to DataFrame
                            if fit_data_df is not None:
                                fit_data_df['Pwr. Law y-fit'] = y_fit_pwr
                        except Exception as err:
                            logging.exception("Scaling Law (Power Law Fit) Error: %s", err, extra={'user': 'SGT Logs'})

                    if opt_gtc["scaling_behavior_truncated_power_law_fit"]["value"] == 1:
                        # 2c. Compute the line of best-fit according to our truncated power-law model
                        try:
                            init_params_cutoff = [1.0, 1.0, 0.1]
                            opt_params_cutoff: np.ndarray = \
                                sp.optimize.curve_fit(truncated_power_law_model, x_avg, y_avg, p0=init_params_cutoff)[0]
                            a_fit_cut, k_fit_cut, c_fit_cut = float(opt_params_cutoff[0]), float(
                                opt_params_cutoff[1]), float(
                                opt_params_cutoff[2])
                            # Generate points for the best-fit curve
                            y_fit_cut = truncated_power_law_model(x_fit, a_fit_cut, k_fit_cut, c_fit_cut)
                            # print(f"Fitted parameters: a={a_fit_cut:.2f}, k={k_fit_cut:.2f}, c={c_fit_cut:.2f}")

                            # 3c. Plot data (truncated power-law best fit)
                            ax, i = plot_axis(i, "Truncated Power Law Fit and Plot of")
                            ax.plot(x_fit, y_fit_cut,
                                    label=f'Fit: $y = ax^{{-k}}*exp(-c*x)$\n$a={a_fit_cut:.2f}, k={k_fit_cut:.2f}, c={c_fit_cut:.2f}$',
                                    color='red')
                            ax.legend()

                            # Write to DataFrame
                            if fit_data_df is not None:
                                fit_data_df['Trunc. Pwr. Law y-fit'] = y_fit_cut
                        except Exception as err:
                            logging.exception("Scaling Law (Truncated Power Law Fit) Error: %s", err,
                                              extra={'user': 'SGT Logs'})

                    if opt_gtc["scaling_behavior_log_normal_fit"]["value"] == 1:
                        # 2d. Compute best-fit, assuming Log-Normal dependence on X
                        try:
                            init_params_log = [1.0, 1.0, 10]
                            opt_params_log: np.ndarray = \
                                sp.optimize.curve_fit(lognormal_model, x_avg, y_avg, p0=init_params_log,
                                                      bounds=([0, 0, 0], [np.inf, np.inf, np.inf]), maxfev=1000)[0]
                            mu_fit, sigma_fit, a_log_fit = float(opt_params_log[0]), float(opt_params_log[1]), float(
                                opt_params_log[2])
                            # Generate predicted points for the best-fit curve
                            y_fit_ln = lognormal_model(x_fit, mu_fit, sigma_fit, a_log_fit)

                            # 3c. Plot data (Log-normal distribution best fit)
                            ax, i = plot_axis(i, "Log-Normal Fit and Plot of")
                            ax.plot(x_fit, y_fit_ln,
                                    label=f'Fit: log-normal shape\n$\\mu={mu_fit:.2f}$, $\\sigma={sigma_fit:.2f}$',
                                    color='red')
                            ax.legend()

                            # Write to DataFrame
                            if fit_data_df is not None:
                                fit_data_df['Log-Normal y-fit'] = y_fit_ln
                        except Exception as err:
                            logging.exception("Scaling Law (Log-Normal Fit) Error: %s", err, extra={'user': 'SGT Logs'})

                # Navigate to the next subplot
                if (i + 1) > 1:
                    if add_plot:
                        plt_figs.append(plt_fig)
                        df_title = f"Nodes-{y_title}"
                        df_title = f"{df_title[:20]}." if len(df_title) > 20 else df_title
                        self._scaling_results[df_title] = data_df.copy()
                        self._scaling_results[f"{df_title} (Fitting)"] = fit_data_df.copy()
                    plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
                    i = 0

            plt_figs.append(plt_fig) if i <= 4 else None

            if best_scale is not None:
                new_row = {'parameter': 'Optimal Image Scale (px)', 'value': int(best_scale)}
                self._results_df = pd.concat([self._results_df, pd.DataFrame([new_row])], ignore_index=True)

            return plt_figs

        def plot_histograms():
            """
            Create plot figures of graph theory histograms selected by the user.

            :return: A list of Matplotlib figures.
            """
            self.update_status([92, "Generating histograms..."])

            opt_gte = graph_obj.configs
            plt_figs = []

            def plot_distribution_histogram(ax: Axes, title: str, distribution: list, x_label: str,
                                            plt_bins: np.ndarray = None, y_label: str = 'Counts'):
                """
                Create a histogram from a distribution dataset.

                :param ax: Plot axis.
                :param title: Title text.
                :param distribution: Dataset to be plotted.
                :param x_label: X-label title text.
                :param plt_bins: Bin dataset.
                :param y_label: Y-label title text.
                :return:
                """
                font_1 = {'fontsize': 9}
                if plt_bins is None:
                    plt_bins = np.linspace(min(distribution), max(distribution), 50)
                try:
                    std_val = str(round(stdev(distribution), 3))
                except StatisticsError:
                    std_val = "N/A"
                hist_title = title + std_val
                ax.set_title(hist_title, fontdict=font_1)
                ax.set(xlabel=x_label, ylabel=y_label)
                ax.hist(distribution, bins=plt_bins)

            # Degree and Closeness
            plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
            if opt_gtc["display_degree_histogram"]["value"] == 1:
                deg_distribution = self._histogram_data["degree_distribution"]
                bins = np.arange(0.5, max(deg_distribution) + 1.5, 1)
                deg_title = r'Degree Distribution: $\sigma$='
                ax_1 = plt_fig.add_subplot(2, 1, 1)
                plot_distribution_histogram(ax_1, deg_title, deg_distribution, 'Degree', plt_bins=bins)

            if opt_gtc["display_closeness_centrality_histogram"]["value"] == 1:
                clo_distribution = self._histogram_data["closeness_distribution"]
                cc_title = r"Closeness Centrality: $\sigma$="
                ax_2 = plt_fig.add_subplot(2, 1, 2)
                plot_distribution_histogram(ax_2, cc_title, clo_distribution, 'Closeness value')
            plt_figs.append(plt_fig)

            # Betweenness, Clustering, Eigenvector and Ohms
            plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
            if opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1:
                bet_distribution = self._histogram_data["betweenness_distribution"]
                bc_title = r"Betweenness Centrality: $\sigma$="
                ax_1 = plt_fig.add_subplot(2, 2, 1)
                plot_distribution_histogram(ax_1, bc_title, bet_distribution, 'Betweenness value')

            if opt_gtc["compute_avg_clustering_coef"]["value"] == 1:
                cluster_coefs = self._histogram_data["clustering_coefficients"]
                clu_title = r"Clustering Coefficients: $\sigma$="
                ax_2 = plt_fig.add_subplot(2, 2, 2)
                plot_distribution_histogram(ax_2, clu_title, cluster_coefs, 'Clust. Coeff.')

            if opt_gtc["display_ohms_histogram"]["value"] == 1:
                ohm_distribution = self._histogram_data["ohms_distribution"]
                oh_title = r"Ohms Centrality: $\sigma$="
                ax_3 = plt_fig.add_subplot(2, 2, 3)
                plot_distribution_histogram(ax_3, oh_title, ohm_distribution, 'Ohms value')

            if opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1:
                eig_distribution = self._histogram_data["eigenvector_distribution"]
                ec_title = r"Eigenvector Centrality: $\sigma$="
                ax_4 = plt_fig.add_subplot(2, 2, 4)
                plot_distribution_histogram(ax_4, ec_title, eig_distribution, 'Eigenvector value')
            plt_figs.append(plt_fig)

            # weighted histograms
            if opt_gte["has_weights"]["value"] == 1:
                wt_type = graph_obj.get_weight_type()
                weight_type = FiberNetworkBuilder.get_weight_options().get(wt_type)

                # degree, betweenness, closeness and eigenvector
                plt_fig = plt.Figure(figsize=(8.5, 11), dpi=300)
                if opt_gtc["display_degree_histogram"]["value"] == 1:
                    w_deg_distribution = self._histogram_data["weighted_degree_distribution"]
                    bins = np.arange(0.5, max(w_deg_distribution) + 1.5, 1)
                    w_deg_title = r"Weighted Degree: $\sigma$="
                    ax_1 = plt_fig.add_subplot(2, 2, 1)
                    plot_distribution_histogram(ax_1, w_deg_title, w_deg_distribution, 'Degree', plt_bins=bins)

                if opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1:
                    w_bet_distribution = self._histogram_data["weighted_betweenness_distribution"]
                    w_bt_title = weight_type + r"-Weighted Betweenness: $\sigma$="
                    ax_2 = plt_fig.add_subplot(2, 2, 2)
                    plot_distribution_histogram(ax_2, w_bt_title, w_bet_distribution, 'Betweenness value')

                if opt_gtc["display_closeness_centrality_histogram"]["value"] == 1:
                    w_clo_distribution = self._histogram_data["weighted_closeness_distribution"]
                    w_clo_title = r"Length-Weighted Closeness: $\sigma$="
                    ax_3 = plt_fig.add_subplot(2, 2, 3)
                    plot_distribution_histogram(ax_3, w_clo_title, w_clo_distribution, 'Closeness value')

                if opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1:
                    w_eig_distribution = self._histogram_data["weighted_eigenvector_distribution"]
                    w_ec_title = weight_type + r"-Weighted Eigenvector Cent.: $\sigma$="
                    ax_4 = plt_fig.add_subplot(2, 2, 4)
                    plot_distribution_histogram(ax_4, w_ec_title, w_eig_distribution, 'Eigenvector value')
                plt_figs.append(plt_fig)

            return plt_figs

        def plot_heatmaps():
            """
            Create plot figures of graph theory heatmaps.

            :return: A list of Matplotlib figures.
            """
            self.update_status([95, "Generating heatmaps..."])

            sz = 30
            lc = 'black'
            plt_figs = []
            opt_gte = graph_obj.configs
            wt_type = graph_obj.get_weight_type()
            weight_type = FiberNetworkBuilder.get_weight_options().get(wt_type)

            def plot_distribution_heatmap(distribution: list, title: str, size: float, line_color: str):
                """
                Create a heatmap from a distribution.

                :param distribution: Dataset to be plotted.
                :param title: Title of the plot figure.
                :param size: Size of the scatter items.
                :param line_color: Color of the line items.
                :return: Histogram plot figure.
                """
                nx_graph = graph_obj.nx_giant_graph
                fig_grp = FiberNetworkBuilder.plot_graph_edges(img_3d, nx_graph, node_distribution_data=distribution,
                                                               plot_nodes=True, edge_color=line_color, node_marker_size=size)

                plt_fig_inner = fig_grp[0]
                plt_fig_inner.set_size_inches(8.5, 8.5)
                plt_fig_inner.set_dpi(400)
                plt_ax = plt_fig_inner.axes[0]
                plt_ax.set_title(title, fontdict={'fontsize': 9})
                plt_ax.set_position([0.05, 0.05, 0.75, 0.75])

                return plt_fig_inner

            if opt_gtc["display_degree_histogram"]["value"] == 1:
                deg_distribution = self._histogram_data["degree_distribution"]
                plt_fig = plot_distribution_heatmap(deg_distribution, 'Degree Heatmap', sz, lc)
                plt_figs.append(plt_fig)
            if (opt_gtc["display_degree_histogram"]["value"] == 1) and (opt_gte["has_weights"]["value"] == 1):
                w_deg_distribution = self._histogram_data["weighted_degree_distribution"]
                plt_title = 'Weighted Degree Heatmap'
                plt_fig = plot_distribution_heatmap(w_deg_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if opt_gtc["compute_avg_clustering_coef"]["value"] == 1:
                cluster_coefs = self._histogram_data["clustering_coefficients"]
                plt_title = 'Clustering Coefficient Heatmap'
                plt_fig = plot_distribution_heatmap(cluster_coefs, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1:
                bet_distribution = self._histogram_data["betweenness_distribution"]
                plt_title = 'Betweenness Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(bet_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if (opt_gtc["display_betweenness_centrality_histogram"]["value"] == 1) and (
                    opt_gte["has_weights"]["value"] == 1):
                w_bet_distribution = self._histogram_data["weighted_betweenness_distribution"]
                plt_title = f'{weight_type}-Weighted Betweenness Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(w_bet_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if opt_gtc["display_closeness_centrality_histogram"]["value"] == 1:
                clo_distribution = self._histogram_data["closeness_distribution"]
                plt_title = 'Closeness Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(clo_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if (opt_gtc["display_closeness_centrality_histogram"]["value"] == 1) and (
                    opt_gte["has_weights"]["value"] == 1):
                w_clo_distribution = self._histogram_data["weighted_closeness_distribution"]
                plt_title = 'Length-Weighted Closeness Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(w_clo_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1:
                eig_distribution = self._histogram_data["eigenvector_distribution"]
                plt_title = 'Eigenvector Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(eig_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if (opt_gtc["display_eigenvector_centrality_histogram"]["value"] == 1) and (
                    opt_gte["has_weights"]["value"] == 1):
                w_eig_distribution = self._histogram_data["weighted_eigenvector_distribution"]
                plt_title = f'{weight_type}-Weighted Eigenvector Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(w_eig_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            if opt_gtc["display_ohms_histogram"]["value"] == 1:
                ohm_distribution = self._histogram_data["ohms_distribution"]
                plt_title = 'Ohms Centrality Heatmap'
                plt_fig = plot_distribution_heatmap(ohm_distribution, plt_title, sz, lc)
                plt_figs.append(plt_fig)
            return plt_figs

        # 1. Compute graphs and plots for scaling behavior
        scaling_figs = plot_scaling_behavior()

        # 2. plotting the original, processed, and binary image, as well as the histogram of pixel grayscale values
        figs = plot_bin_images()
        for fig in figs:
            out_figs.append(fig)

        # 3a. plotting graph nodes
        fig = graph_obj.plot_graph_network(image_arr=img_3d, plot_nodes=True, a4_size=True)
        if fig is not None:
            out_figs.append(fig)

        # 4b. plotting graph edges
        fig = graph_obj.plot_graph_network(image_arr=img_3d, a4_size=True)
        if fig is not None:
            out_figs.append(fig)

        # 4a. displaying all the GT calculations in Table (on the entire page)
        fig, fig_wt = plot_gt_results()
        out_figs.append(fig)
        if fig_wt:
            out_figs.append(fig_wt)

        # 4b. display scaling GT results in a Table
        for fig in scaling_figs:
            out_figs.append(fig)

        # 5a. displaying histograms
        figs = plot_histograms()
        for fig in figs:
            out_figs.append(fig)

        # 5b. displaying heatmaps
        if opt_gtc["display_heatmaps"]["value"] == 1:
            figs = plot_heatmaps()
            for fig in figs:
                out_figs.append(fig)

        # 6. displaying run information
        fig = plot_run_configs()
        out_figs.append(fig)
        return out_figs

    @staticmethod
    def write_to_pdf(sgt_obj, update_func=None) -> bool:
        """
        Write results to a PDF file.

        Args:
            sgt_obj: StructuralGT object with calculated GT parameters
            update_func: Callable for progress updates (e.g., update_func(percentage, message))

        Returns:
            True if the PDF file is written successfully, otherwise False
        """
        try:
            if update_func:
                update_func(98, "Writing PDF...")

            filename, output_location = sgt_obj.ntwk_p.get_filenames()
            pdf_filename = filename + "_SGT_results.pdf"
            pdf_file = os.path.join(output_location, pdf_filename)

            if not sgt_obj.plot_figures:
                raise ValueError("No figures available to write to PDF.")

            with PdfPages(pdf_file) as pdf:
                for fig in sgt_obj.plot_figures:
                    pdf.savefig(fig)

            if sgt_obj.results_df is not None:
                csv_filename = filename + "_SGT_unweighted.csv"
                csv_file = os.path.join(output_location, csv_filename)
                sgt_obj.results_df.to_csv(csv_file, index=False)

            if sgt_obj.weighted_results_df is not None:
                csv_filename = filename + "_SGT_weighted.csv"
                csv_file = os.path.join(output_location, csv_filename)
                sgt_obj.weighted_results_df.to_csv(csv_file, index=False)

            if sgt_obj.scaling_results:
                excel_filename = filename + "_SGT_scaling.xlsx"
                excel_file = os.path.join(output_location, excel_filename)
                with pd.ExcelWriter(str(excel_file), engine='xlsxwriter') as writer:
                    for tbl_title, tbl_df in sgt_obj.scaling_results.items():
                        # Clean sheet name: Excel allows max 31 chars, no : \ / ? * [ ]
                        safe_title = str(tbl_title)[:31].replace(":", "").replace("\\", "").replace("/", "") \
                            .replace("?", "").replace("*", "").replace("[", "").replace("]", "")
                        tbl_df.to_excel(writer, sheet_name=safe_title, index=False)
            return True
        except Exception as err:
            logging.exception("GT Computation Error: %s", err, extra={'user': 'SGT Logs'})
            if update_func:
                update_func(-1, "Error occurred while trying to write to PDF.")
            return False

    @staticmethod
    def safe_run_analyzer(sgt_obj, update_func, save_to_pdf=False) -> tuple[bool, None] | tuple[bool, "GraphAnalyzer"]:
        """
        Safely compute GT descriptors without raising exceptions or crushing app.

        Args:
            sgt_obj: StructuralGT object with calculated GT parameters
            update_func: Callable for progress updates (e.g., update_func(percentage, message))
            save_to_pdf: Save results to a PDF file
        """
        try:
            # Add Listeners
            sgt_obj.add_listener(update_func)

            # Run GT computations
            sgt_obj.run_analyzer()
            if sgt_obj.abort:
                raise AbortException("Process aborted")

            # Write GT results to PDF
            if save_to_pdf:
                GraphAnalyzer.write_to_pdf(sgt_obj, update_func)

            # Cleanup - remove listeners
            sgt_obj.remove_listener(update_func)
            return True, sgt_obj
        except AbortException:
            update_func(-1, "Task aborted by user or a fatal error occurred!")
            sgt_obj.remove_listener(update_func)
            return False, None
        except Exception as err:
            update_func(-1, "Error encountered! Try again")
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            # Clean up listeners before exiting
            sgt_obj.remove_listener(update_func)
            return False, None

    @staticmethod
    def safe_run_multi_analyzer(sgt_objs, update_func) -> None|dict:
        """
        Safely compute GT descriptors of multiple images without raising exceptions or crushing the app.

        Args:
            sgt_objs: List of StructuralGT objects with calculated GT parameters
            update_func: Callable for progress updates (e.g., update_func(percentage, message))
        """
        try:
            i = 0
            keys_list = list(sgt_objs.keys())
            for key in keys_list:
                sgt_obj = sgt_objs[key]

                status_msg = f"Analyzing Image: {(i + 1)} / {len(sgt_objs)}"
                update_func(101, status_msg)

                start = time.time()
                success, new_sgt = GraphAnalyzer.safe_run_analyzer(sgt_obj, update_func)
                # TerminalApp.is_aborted(sgt_obj)
                if success:
                    GraphAnalyzer.write_to_pdf(new_sgt, update_func)
                end = time.time()

                i += 1
                output = status_msg + "\n" + f"Run-time: {str(end - start)}  seconds\n"
                output += "Results generated for: " + sgt_obj.ntwk_p.img_path + "\n"
                # graph_obj = sgt_obj.ntwk_p.graph_obj
                # output += "Node Count: " + str(graph_obj.nx_giant_graph.number_of_nodes()) + "\n"
                # output += "Edge Count: " + str(graph_obj.nx_giant_graph.number_of_edges()) + "\n"
                # filename, out_dir = sgt_obj.ntwk_p.get_filenames()
                # out_file = os.path.join(out_dir, filename + '-v2_results.txt')
                # write_txt_file(output, out_file)
                logging.info(output, extra={'user': 'SGT Logs'})
            return sgt_objs
        except AbortException:
            update_func(-1, "Task aborted by user or a fatal error occurred!")
            return None
        except Exception as err:
            update_func(-1, "Error encountered! Try again")
            logging.exception("Error: %s", err, extra={'user': 'SGT Logs'})
            return None



# SPDX-License-Identifier: GNU GPL v3

"""
Builds a graph network from nanoscale microscopy images.
"""

import os
import numbers
import itertools
import numpy as np
import pandas as pd
import igraph as ig
import networkx as nx
import matplotlib.pyplot as plt
from cv2.typing import MatLike

from .sknw_mod import build_sknw#, build_graph
from ..utils.progress_update import ProgressUpdate
from ..networks.graph_skeleton import GraphSkeleton
from ..utils.config_loader import load_gte_configs
from ..utils.sgt_utils import write_gsd_file, gsd_to_skeleton, csv_to_graph


class FiberNetworkBuilder(ProgressUpdate):
    """
    A class for builds a graph network from microscopy images and stores is as a NetworkX object.

    """

    def __init__(self, cfg_file=""):
        """
        A class for builds a graph network from microscopy images and stores is as a NetworkX object.

        Args:
            cfg_file (str): configuration file path

        """
        super(FiberNetworkBuilder, self).__init__()
        self._configs: dict = load_gte_configs(cfg_file)  # graph extraction parameters and options.
        self._props: list = []
        self._img_ntwk: MatLike | None = None
        self._nx_giant_graph: nx.Graph | None = None
        self._nx_graph: nx.Graph | None = None
        self._ig_graph: None | ig.Graph = None
        self._gsd_file: str | None = None
        self._skel_obj: GraphSkeleton | None = None

    @property
    def configs(self) -> dict:
        """Returns the graph extraction configuration parameters and options."""
        return self._configs

    @configs.setter
    def configs(self, configs: dict):
        """Sets the graph extraction configuration parameters and options."""
        self._configs = configs

    @property
    def props(self) -> list:
        """Returns the computed graph properties as a list-array."""
        return self._props

    @property
    def img_ntwk(self) -> MatLike | None:
        """Returns the processed image with the graph drawn on it."""
        return self._img_ntwk

    @img_ntwk.setter
    def img_ntwk(self, img_ntwk: MatLike | None):
        """Sets the processed image with the graph drawn on it."""
        self._img_ntwk = img_ntwk

    @property
    def nx_giant_graph(self) -> nx.Graph | None:
        """Returns the giant graph of the NetworkX object."""
        return self._nx_giant_graph

    @nx_giant_graph.setter
    def nx_giant_graph(self, nx_giant_graph: nx.Graph | None):
        """Sets the giant graph of the NetworkX object."""
        self._nx_giant_graph = nx_giant_graph

    @property
    def nx_graph(self) -> nx.Graph | None:
        """Returns the NetworkX graph object."""
        return self._nx_graph

    @nx_graph.setter
    def nx_graph(self, nx_graph: nx.Graph | None):
        """Sets the NetworkX graph object."""
        self._nx_graph = nx_graph

    @property
    def ig_graph(self) -> None | ig.Graph:
        """Returns the iGraph graph object."""
        return self._ig_graph

    @property
    def gsd_file(self) -> str | None:
        """Returns the filename of the graph skeleton saved in GSD format."""
        return self._gsd_file

    @property
    def skel_obj(self):
        return self._skel_obj

    def fit_graph(self, save_dir: str, input_data: MatLike | str = None, is_img_2d: bool = True, px_width_sz: float = 1.0, rho_val: float = 1.0, file_name: str = "img") -> None:
        """
        Execute functions that build a NetworkX graph from the binary image.

        :param save_dir: Directory to save the graph to.
        :param input_data: A binary image for building Graph Skeleton for the NetworkX graph OR a file path for loading the NetworkX graph from a file.
        :param is_img_2d: Whether the image is 2D or 3D otherwise.
        :param px_width_sz: Width of a pixel in nanometers.
        :param rho_val: Resistivity coefficient/value of the material.
        :param file_name: Filename of the binary image.
        :return:
        """

        if self.abort:
            self.update_status([-1, "Task aborted by due to an error. If problem with graph: change/apply different "
                                    "image/binary filters and graph options. OR change brightness/contrast"])
            return

        if type(input_data) is str:
            self.update_status([50, "Loading graph network from file..."])
            nx_graph = self.create_graph_from_file(input_data)
        elif type(input_data) is np.ndarray:
            self.update_status([50, "Extracting the graph network..."])
            nx_graph = self.extract_graph(image_bin=input_data, is_img_2d=is_img_2d, px_size=px_width_sz, rho_val=rho_val)
        else:
            self.update_status([-1, "Invalid input for building a graph network. Either provide a graph file path or a binary image as an OpenCV matrix."])
            self.abort = True
            return

        self.update_status([70, "Verifying graph network..."])
        success = self.verify_graph(nx_graph)
        if not success:
            self.update_status([-1, "Problem encountered, change image/binary filters and graph options. OR change brightness/contrast"])
            self.abort = True
            return

        self.update_status([77, "Retrieving graph properties..."])
        self._props = self.get_graph_props()

        self.update_status([90, "Saving graph network..."])
        # Save graph to GSD/HOOMD - For OVITO rendering
        self._configs["export_as_gsd"]["value"] = 1
        self.save_graph_to_file(file_name, save_dir)

    def reset_graph(self) -> None:
        """
        Erase the existing data stored in the object.
        :return:
        """
        self.nx_graph, self._ig_graph, self._img_ntwk = None, None, None

    def verify_graph(self, nx_graph) -> bool:
        """
        Verify if the NetworkX graph is valid. If it is valid, save in object members.

        :param nx_graph: The NetworkX graph to verify.
        :return: True if the graph is valid, False otherwise.
        """
        if nx_graph is None:
            return False

        if nx_graph.number_of_edges() <= 0 or nx_graph.number_of_nodes() <= 0:
            return False

        # Save NetworkX graph
        self.nx_graph = nx_graph
        # Save iGraph graph
        self._ig_graph = ig.Graph.from_networkx(nx_graph)
        # Save giant NetworkX graph
        connected_components = list(nx.connected_components(nx_graph))
        if not connected_components:  # In case the graph is empty
            connected_components = []
        sub_graphs = [nx_graph.subgraph(c).copy() for c in connected_components]
        if sub_graphs:
            giant_graph = max(sub_graphs, key=lambda g: g.number_of_nodes())
        else:
            giant_graph = nx_graph
        self._nx_giant_graph = giant_graph
        return True

    def extract_graph(self, image_bin: MatLike = None, is_img_2d: bool = True, px_size: float = 1.0, rho_val: float = 1.0) -> nx.Graph | None:
        """
        Build a skeleton from the image and use the skeleton to build a NetworkX graph.

        :param image_bin: Binary image from which the skeleton will be built and graph drawn.
        :param is_img_2d: Whether the image is 2D or 3D otherwise.
        :param px_size: Width of a pixel in nanometers.
        :param rho_val: Resistivity coefficient/value of the material.
        :return:
        """
        try:
            if image_bin is None:
                return None

            opt_gte = self._configs
            if opt_gte is None:
                return None

            self.update_status([51, "Building skeleton from binary image..."])
            graph_skel = GraphSkeleton(image_bin, opt_gte, is_2d=is_img_2d, progress_func=self.update_status)
            self._skel_obj = graph_skel
            img_skel = graph_skel.skeleton

            self.update_status([60, "Creating graph network..."])
            # nx_graph = sknw.build_sknw(img_skel)
            nx_graph = build_sknw(img_skel)

            if opt_gte["remove_self_loops"]["value"]:
                self.update_status([64, "Removing self loops from graph network..."])

            self.update_status([66, "Assigning weights to graph network..."])
            for (s, e) in list(nx_graph.edges()):
                if opt_gte["remove_self_loops"]["value"]:
                    # Removing all instances of edges where the start and end are the same, or "self-loops"
                    if s == e:
                        nx_graph.remove_edge(s, e)
                        continue

                # 'sknw' library stores the length of edge as 'weight', we create an attribute 'length', and update 'weight'
                nx_graph[s][e]['length'] = nx_graph[s][e]['weight']
                ge = nx_graph[s][e]['pts']

                if opt_gte["has_weights"]["value"] == 1:
                    # We update 'weight'
                    wt_type = self.get_weight_type()
                    weight_options = FiberNetworkBuilder.get_weight_options()
                    pix_width, pix_angle, wt = graph_skel.assign_weights(ge, wt_type, weight_options=weight_options,
                                                                             pixel_dim=px_size, rho_dim=rho_val)
                    nx_graph[s][e]['width'] = pix_width
                    nx_graph[s][e]['angle'] = pix_angle
                    nx_graph[s][e]['weight'] = wt
                else:
                    pix_width, pix_angle, wt = graph_skel.assign_weights(ge, None)
                    nx_graph[s][e]['width'] = pix_width
                    nx_graph[s][e]['angle'] = pix_angle
                    del nx_graph[s][e]['weight']            # delete 'weight'
                # print(f"{nx_graph[s][e]}\n")

            return nx_graph
        except Exception as e:
            self.update_status([-1, f"Problem encountered while extracting graph from binary image: {e}"])
            return None

    def create_graph_from_file(self, file_path: str) -> nx.Graph | None:
        """
        Load a NetworkX graph from a file that may contain:
          - Edge list (2 columns)
          - Adjacency matrix (square matrix)
          - XYZ positions (3 columns: x, y, z, edges inferred by distance threshold)

        :param file_path: Path to the graph file
        :return: True if the graph is read, False otherwise
        """
        try:
            self.update_status([60, "Reading graph network..."])
            ext = os.path.splitext(file_path)[1].lower()
            if ext == ".gsd":
                skel = gsd_to_skeleton(file_path)
                self._skel_obj = GraphSkeleton(np.array([None]))
                self._skel_obj._skeleton = skel
                # self._skel_obj._skeleton_3d = np.asarray([skel])
                nx_graph = build_sknw(skel)
                for (s, e) in list(nx_graph.edges()):
                    nx_graph[s][e]['length'] = nx_graph[s][e]['weight']
                    nx_graph[s][e]['width'] = 1.0
            elif ext == ".csv":
                nx_graph = csv_to_graph(file_path)
                # self._skel_obj = GraphSkeleton(np.array([None]))
                # self._skel_obj._skeleton = skel
                ## self._skel_obj._skeleton_3d = np.asarray([skel])
                # nx_graph = build_sknw(skel)
            else:
                self.update_status([-1, f"Unsupported file extension: {ext}. If CSV, comma as delimiter."])
                nx_graph = None
            return nx_graph
        except Exception as e:
            self.update_status([-1, f"Problem encountered while loading graph from file: {e}"])
            return None

    def plot_graph_network(self, image_arr: MatLike, giant_only: bool = False, plot_nodes: bool = False, a4_size: bool = False) -> None | plt.Figure:
        """
        Creates a plot figure of the graph network. It draws all the edges and nodes of the graph.

        :param image_arr: Slides of 2D images to be used to draw the network.
        :param giant_only: If True, only the giant graph is identified and drawn.
        :param plot_nodes: Make the graph's node plot figure.
        :param a4_size: Decision if to create an A4 size plot figure.

        :return:
        """

        if self.nx_graph is None:
            return None

        # Fetch the graph and config options
        if giant_only:
            nx_graph = self._nx_giant_graph
        else:
            nx_graph = self.nx_graph
        show_node_id = (self._configs["display_node_id"]["value"] == 1)
        add_width_thickness = (self._configs["add_width_thickness"]["value"] == 1)

        # Create the plot figure(s)
        fig_grp = FiberNetworkBuilder.plot_graph_edges(image_arr, nx_graph, plot_nodes=plot_nodes, show_node_id=show_node_id, add_width_thickness=add_width_thickness)
        fig = fig_grp[0]
        if a4_size:
            plt_title = "Graph Node Plot" if plot_nodes else "Graph Edge Plot"
            fig.set_size_inches(8.5, 11)
            fig.set_dpi(400)
            ax = fig.axes[0]
            ax.set_title(plt_title)
            # This moves the Axes to start: 5% from the left, 5% from the bottom,
            # and have a width and height: 80% of the figure.
            # [left, bottom, width, height]
            ax.set_position([0.05, 0.05, 0.9, 0.9])
        return fig

    def get_config_info(self) -> str:
        """
        Get the user selected parameters and options information.
        :return:
        """

        opt_gte = self._configs

        run_info = "***Graph Extraction Configurations***\n"
        if opt_gte["has_weights"]["value"] == 1:
            wt_type = self.get_weight_type()
            run_info += f"Weight Type: {FiberNetworkBuilder.get_weight_options().get(wt_type)} || "
        if opt_gte["merge_nearby_nodes"]["value"]:
            run_info += "Merge Nodes || "
        if opt_gte["prune_dangling_edges"]["value"]:
            run_info += "Prune Dangling Edges || "
        run_info = run_info[:-3] + '' if run_info.endswith('|| ') else run_info
        run_info += "\n"
        if opt_gte["remove_disconnected_segments"]["value"]:
            run_info += f"Remove Objects of Size = {opt_gte["remove_disconnected_segments"]["items"][0]["value"]} || "
        if opt_gte["remove_self_loops"]["value"]:
            run_info += "Remove Self Loops || "
        run_info = run_info[:-3] + '' if run_info.endswith('|| ') else run_info

        return run_info

    def get_graph_props(self) -> list:
        """
        A method that retrieves graph properties and stores them in a list-array.

        Returns: list of graph properties
        """

        # 1. Identify the subcomponents (graph segments) that make up the entire NetworkX graph.
        self.update_status([78, "Identifying graph subcomponents..."])
        graph = self.nx_graph.copy()
        connected_components = list(nx.connected_components(graph))
        if not connected_components:  # In case the graph is empty
            connected_components = []
        sub_graphs = [graph.subgraph(c).copy() for c in connected_components]
        num_graphs = len(sub_graphs)
        connect_ratio = self._nx_giant_graph.number_of_nodes() / graph.number_of_nodes()

        # 2. Populate graph properties
        self.update_status([80, "Storing graph properties..."])
        props = [
            ["Weight Type", str(FiberNetworkBuilder.get_weight_options().get(self.get_weight_type()))],
            ["Edge Count", str(graph.number_of_edges())],
            ["Node Count", str(graph.number_of_nodes())],
            ["Graph Count", str(len(connected_components))],
            ["Sub-graph Count", str(num_graphs)],
            ["Giant graph ratio", f"{round((connect_ratio * 100), 3)}%"]]
        return props

    def get_weight_type(self) -> str | None:
        wt_type = None  # Default weight
        if self._configs["has_weights"]["value"] == 0:
            return wt_type

        for i in range(len(self._configs["has_weights"]["items"])):
            if self._configs["has_weights"]["items"][i]["value"]:
                wt_type = self._configs["has_weights"]["items"][i]["id"]
        return wt_type

    def save_graph_to_file(self, filename: str, out_dir: str) -> None:
        """
        Save graph data into files.

        :param filename: The filename to save the data to.
        :param out_dir: The directory to save the data to.
        :return:
        """

        nx_graph = self.nx_graph.copy()
        opt_gte = self._configs

        g_filename = filename + "_graph.gexf"
        edges_filename = filename + "_EdgeList.csv"
        nodes_filename = filename + "_NodePositions.csv"
        adj_filename = filename + "_AdjMat.csv"
        gsd_filename = filename + "_skel.gsd"
        gexf_file = os.path.join(out_dir, g_filename)
        edges_file = os.path.join(out_dir, edges_filename)
        nodes_file = os.path.join(out_dir, nodes_filename)
        adj_file = os.path.join(out_dir, adj_filename)

        if opt_gte["export_adj_mat"]["value"] == 1:
            adj_mat = nx.adjacency_matrix(self.nx_graph).todense()
            np.savetxt(str(adj_file), adj_mat, delimiter=",")

        if opt_gte["export_edge_list"]["value"] == 1:
            if opt_gte["has_weights"]["value"] == 1:
                cols = ['Source', 'Target', 'Weight', 'Length', 'Width', 'Angle']
                lst_edges = []
                for (s, e) in list(nx_graph.edges()):
                    weight = (nx_graph[s][e]['weight'])
                    length = nx_graph[s][e]['length']
                    width = nx_graph[s][e]['width']
                    angle = nx_graph[s][e]['angle']
                    el = [s, e, weight, length, width, angle]
                    lst_edges.append(el)
                df_edges = pd.DataFrame(lst_edges, columns=cols)
                df_edges.to_csv(edges_file, index=False)
            else:
                cols = ['Source', 'Target']
                lst_edges = []
                for (s, e) in list(nx_graph.edges()):
                    el = [s, e]
                    lst_edges.append(el)
                df_edges = pd.DataFrame(lst_edges, columns=cols)
                df_edges.to_csv(edges_file, index=False)

        if opt_gte["export_node_positions"]["value"] == 1:
            node_list = list(nx_graph.nodes())
            node_pos = np.array([nx_graph.nodes[i]['o'] for i in node_list])
            df_node_pos = pd.DataFrame(node_pos, columns=['x', 'y'])
            df_node_pos.to_csv(nodes_file, index=False)

        if opt_gte["export_as_gexf"]["value"] == 1:
            # deleting extraneous info and then exporting the final skeleton
            for (x) in nx_graph.nodes():
                del nx_graph.nodes[x]['pts']
                del nx_graph.nodes[x]['o']
            for (s, e) in nx_graph.edges():
                del nx_graph[s][e]['pts']
            nx.write_gexf(nx_graph, gexf_file)

        if opt_gte["export_as_gsd"]["value"] == 1:
            self._gsd_file = os.path.join(out_dir, gsd_filename)
            if self._skel_obj is not None:
                if self._skel_obj.skeleton_3d is not None:
                    write_gsd_file(self._gsd_file, self._skel_obj.skeleton_3d)

    @staticmethod
    def get_weight_options() -> dict:
        """
        Returns the weight options for building the graph edges.

        :return:
        """
        weight_options = {
            'DIA': 'Diameter',
            'AREA': 'Area',  # surface area of edge
            'LEN': 'Length',
            'ANGLE': 'Angle',
            'INV_LEN': 'InverseLength',
            'VAR_CON': 'Conductance',  # with variable width
            'FIX_CON': 'FixedWidthConductance',
            'RES': 'Resistance',
            # '': ''
        }
        return weight_options

    @staticmethod
    def plot_graph_edges(image: MatLike, nx_graph: nx.Graph, node_distribution_data: list = None, plot_nodes: bool = False, show_node_id: bool = False, add_width_thickness: bool = False, transparent: bool = False, edge_color: str= 'r', node_marker_size: float = 3) -> dict:
        """
        Plot graph edges on top of the image

        :param image: image to be superimposed with graph edges;
        :param nx_graph: a NetworkX graph;
        :param node_distribution_data: a list of node distribution data for a heatmap plot;
        :param plot_nodes: whether to plot graph nodes or not;
        :param show_node_id: if True, node IDs are displayed on the plot;
        :param add_width_thickness: whether to add width thickness to node distribution data;
        :param transparent: whether to draw the image with a transparent background;
        :param edge_color: each edge's line color;
        :param node_marker_size: the size (diameter) of the node marker
        :return:
        """

        def plot_graph_nodes(node_ax):
            """
            Plot graph nodes on top of the image.
            :param node_ax: Matplotlib axes
            """

            node_list = list(nx_graph.nodes())
            gn = np.array([nx_graph.nodes[i]['o'] for i in node_list])

            if show_node_id:
                i = 0
                for x, y in zip(gn[:, coord_1], gn[:, coord_2]):
                    node_ax.annotate(str(i), (x, y), fontsize=5)
                    i += 1

            if node_distribution_data is not None:
                c_set = node_ax.scatter(gn[:, coord_1], gn[:, coord_2], s=node_marker_size, c=node_distribution_data, cmap='plasma')
                return c_set
            else:
                # c_set = node_ax.scatter(gn[:, coord_1], gn[:, coord_2], s=marker_size)
                node_ax.plot(gn[:, coord_1], gn[:, coord_2], 'b.', markersize=node_marker_size)
                return None

        def create_plt_axes(pos) -> plt.Figure:
            """
            Create a matplotlib axes object.
            Args:
                pos: index position of image frame.

            Returns:

            """
            new_fig = plt.Figure()
            new_ax = new_fig.add_axes((0, 0, 1, 1))  # span the whole figure
            new_ax.set_axis_off()

            if image is None:
                return new_fig

            if transparent:
                new_ax.imshow(image[pos], cmap='gray', alpha=0)  # Alpha=0 makes image 100% transparent
            else:
                new_ax.imshow(image[pos], cmap='gray')
            return new_fig

        def normalize_width(w:float|numbers.Real|dict, new_min=0.5, new_max=5.0) -> float:
            if max_w == min_w:
                return (new_min + new_max) / 2  # avoid division by zero
            if not isinstance(w, numbers.Real):
                print(f"Invalid width type ({type(w)}); using default normalized width = 1.0")
                return 1.0
            norm_w = new_min + (w - min_w) * (new_max - new_min) / (max_w - min_w)
            return float(norm_w)

        def get_max_dims():
            xs, ys = [], []
            node_list = list(nx_graph.nodes())
            for i in node_list:
                pts = nx_graph.nodes[i].get("pts", None)
                if pts is not None:
                    pts = np.array(pts)
                    if pts.ndim == 1:  # single point [x, y]
                        xs.append(pts[0])
                        ys.append(pts[1])
                    elif pts.ndim == 2:  # multiple points [[x,y], [x,y], ...]
                        xs.extend(pts[:, 0])
                        ys.extend(pts[:, 1])

            if not xs or not ys:
                return None, None  # no "pts" found

            max_x = max(xs)
            max_y = max(ys)
            return max_x, max_y

        fig_group = {0: create_plt_axes(0)}
        if image is None:
            img_w, img_h = get_max_dims()
            if img_w is not None and img_h is not None:
                # GSD file has image dimensions but no image data
                image = np.ones((img_w, img_h), dtype=np.uint8) * 255
                image = [image]
            else:
                # Draw graph using NetworkX library
                ax = fig_group[0].get_axes()[0]
                if node_distribution_data is None:
                    # Planar: tries to avoid edge crossings, (working only for planar graphs?)
                    nx.draw(nx_graph, ax=ax, with_labels=show_node_id, node_size=node_marker_size, edge_color=edge_color)
                else:
                    # Normalize values for colormap
                    v_min, v_max = min(node_distribution_data), max(node_distribution_data)
                    nx.draw(nx_graph, ax=ax, with_labels=show_node_id,
                                   node_size=node_marker_size, node_color=node_distribution_data, cmap='plasma',
                                   vmin=v_min, vmax=v_max, edge_color=edge_color)
                    # Add colorbar for heatmap
                    sm = plt.cm.ScalarMappable(cmap='plasma', norm=plt.Normalize(vmin=v_min, vmax=v_max))
                    sm.set_array([])  # required for colorbar
                    cbar = fig_group[0].colorbar(sm, ax=ax, orientation='vertical', label='Value')
                    cbar.ax.set_position([0.82, 0.05, 0.05, 0.9])
                return fig_group

        # First, extract all widths to compute min and max
        all_widths = np.array([nx_graph[s][e]['width'] for s, e in nx_graph.edges()])
        if all_widths.size == 0:
            return fig_group
        min_w, max_w = min(all_widths), max(all_widths)

        # Create a color cycle for each graph component
        if edge_color == 'black':
            color_list = ['k', 'r', 'g', 'b', 'c', 'm', 'y']
        else:
            color_list = ['r', 'y', 'g', 'b', 'c', 'm', 'k']
        color_cycle = itertools.cycle(color_list)
        nx_components = list(nx.connected_components(nx_graph))
        for component in nx_components:
            color = next(color_cycle)
            sg = nx_graph.subgraph(component)

            for (s, e) in sg.edges():
                ge = sg[s][e]['pts']
                edge_w = 0.8
                if add_width_thickness:
                    wt = sg[s][e]['width']
                    edge_w = normalize_width(wt)  # The size of the plot line-width depends on width of edge
                coord_1, coord_2 = 1, 0  # coordinates: (y, x)
                coord_3 = 0
                if np.array(ge).shape[1] == 3:
                    # image and graph are 3D (not 2D)
                    # 3D Coordinates are (x, y, z) ... assume that y and z are the same for 2D graphs and x is depth.
                    coord_1, coord_2, coord_3 = 2, 1, 0  # coordinates: (z, y, x)

                if coord_3 in fig_group and fig_group[coord_3] is not None:
                    pass
                else:
                    fig_group[coord_3] = create_plt_axes(coord_3)
                ax = fig_group[coord_3].get_axes()[0]
                ax.plot(ge[:, coord_1], ge[:, coord_2], color, linewidth=edge_w)

        if plot_nodes:
            for idx, plt_fig in fig_group.items():
                ax = plt_fig.get_axes()[0]
                node_color_set = plot_graph_nodes(ax)
                if node_color_set is not None:
                    cbar = plt_fig.colorbar(node_color_set, ax=ax, orientation='vertical', label='Value')
                    # [left, bottom, width, height]
                    cbar.ax.set_position([0.82, 0.05, 0.05, 0.9])
        return fig_group



from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt



# Define a simple table model
class TableModel(QAbstractTableModel):
    SelectedRole = Qt.ItemDataRole.UserRole + 20

    def __init__(self, data, parent=None):
        super().__init__(parent)
        self.itemData = data
        self.imageCache = {}
        self.selected_index = -1  # To track the selected row

    def rowCount(self, parent=QModelIndex()):
        return len(self.itemData) if self.itemData else 0

    def columnCount(self, parent=QModelIndex()):
        return len(self.itemData[0]) if self.itemData else 0

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if not index.isValid() or not (0 <= index.row() < self.rowCount()):
            return None
        if role == Qt.ItemDataRole.DisplayRole:
            return self.itemData[index.row()][index.column()]
        elif role == Qt.ItemDataRole.DecorationRole:
            if len(self.imageCache) <= 0:
                return None
            image_name = self.itemData[index.row()][index.column()]
            return self.imageCache[image_name]
        elif role == self.SelectedRole:
            return index.row() == self.selected_index  # True if selected
        return None

    def set_selected(self, row):
        if 0 <= row < len(self.itemData):
            old_index = self.selected_index
            self.selected_index = row
            if old_index != -1:
                self.dataChanged.emit(self.index(old_index, 0), self.index(old_index, 0), [self.SelectedRole])
            self.dataChanged.emit(self.index(row, 0), self.index(row, 0), [self.SelectedRole])

    def reset_data(self, item_data):
        self.beginResetModel()
        self.imageCache = {}
        self.itemData = item_data
        self.endResetModel()
        self.dataChanged.emit(self.index(0, 0), self.index(len(self.itemData) - 1, 0))

    def update_data(self, img_list, img_cache):
        """Updates model with new images from analyze_objs"""
        self.beginResetModel()
        if img_list is None or img_cache is None:
            # No data to add/update
            self.itemData = []
            self.imageCache = {}
            self.endResetModel()
            # Emit dataChanged signal to notify QML
            self.dataChanged.emit(self.index(0, 0), self.index(len(self.itemData) - 1, 0))
            return

        start_row = len(self.itemData)
        self.itemData = img_list
        self.imageCache = img_cache
        self.endResetModel()
        # Emit dataChanged signal to notify QML
        self.dataChanged.emit(self.index(start_row, 0), self.index(len(self.itemData) - 1, 0))

    def roleNames(self):
        return {
            Qt.ItemDataRole.DisplayRole: b"text",
            Qt.ItemDataRole.DecorationRole: b"thumbnail",
            self.SelectedRole: b"selected",
        }

    def headerData(self, section, orientation, role=Qt.ItemDataRole.DisplayRole):
         if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            return f"Column {section + 1}"
         return super().headerData(section, orientation, role)